from typing import Tuple, Dict, List
import itertools
import numpy as np
from panqec.codes import StabilizerCode

Operator = Dict[Tuple, str]  # Location to pauli ('X','Y','Z')
Coordinates = List[Tuple]  # List of locations


class Color3DCode(StabilizerCode):
    """3D Color Code on periodic truncated octahedral lattice.

    Parameters
    ----------
    L_x : int
        Size in the x direction.
    L_y : int
        Size in the y direction.
    L_z : int
        Size in the z direction.

    Notes
    -----
    See
    `Bombin and Martin-Delgado 2007 <https://arxiv.org/abs/cond-mat/0607736>`_
    for more information.
    """
    dimension = 3

    @property
    def label(self) -> str:
        return 'Color {}x{}x{}'.format(*self.size)

    def get_qubit_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        coordinates = []

        stab_coordinates = self.get_stabilizer_coordinates()

        for location in stab_coordinates:
            qubit_coords = list(self.get_stabilizer(location).keys())
            for coord in qubit_coords:
                if coord not in coordinates:
                    coordinates.append(coord)

        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        # Yellow and red cells
        for z in range(2, 4*Lz, 4):
            for x in range(2, 4*Lx, 4):
                for y in range(2, 4*Ly, 4):
                    coordinates.append((x, y, z))

        # Blue and green cells
        for z in range(4, 4*Lz+1, 4):
            for x in range(4, 4*Lx+1, 4):
                for y in range(4, 4*Ly+1, 4):
                    coordinates.append((x, y, z))

        # Yellow and red square faces orthogonal to z axis
        for z in range(0, 4*Lz, 4):
            for x in range(2, 4*Lx, 4):
                for y in range(2, 4*Ly, 4):
                    coordinates.append((x, y, z))

        # Yellow and red square faces orthogonal to x axis
        for z in range(2, 4*Lz, 4):
            for x in range(0, 4*Lx, 4):
                for y in range(2, 4*Ly, 4):
                    coordinates.append((x, y, z))

        # Yellow and red square faces orthogonal to y axis
        for z in range(2, 4*Lz, 4):
            for x in range(2, 4*Lx, 4):
                for y in range(0, 4*Ly, 4):
                    coordinates.append((x, y, z))

        # Blue and green square faces orthogonal to z axis
        for z in range(2, 4*Lz+1, 4):
            for x in range(4, 4*Lx+1, 4):
                for y in range(4, 4*Ly+1, 4):
                    coordinates.append((x, y, z))

        # Blue and green square faces orthogonal to x axis
        for z in range(4, 4*Lz+1, 4):
            for x in range(2, 4*Lx+1, 4):
                for y in range(4, 4*Ly+1, 4):
                    coordinates.append((x, y, z))

        # Blue and green square faces orthogonal to y axis
        for z in range(4, 4*Lz+1, 4):
            for x in range(4, 4*Lx+1, 4):
                for y in range(2, 4*Ly+1, 4):
                    coordinates.append((x, y, z))

        # All hexagons
        for z in range(1, 4*Lz+2, 2):
            for x in range(1, 4*Lx+2, 2):
                for y in range(1, 4*Ly+2, 2):
                    coordinates.append((x, y, z))

        return coordinates

    def stabilizer_type(self, location: Tuple) -> str:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        x, y, z = location

        cell_color = {6: 'cell-yellow',
                      2: 'cell-red',
                      4: 'cell-green',
                      0: 'cell-blue'}

        if x % 2 == 1:
            stab_type = 'face-hex'
        elif x % 4 == z % 4 and y % 4 == z % 4:
            stab_type = cell_color[(x + y + z) % 8]
        else:
            stab_type = 'face-square'

        return stab_type

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        Lx, Ly, Lz = self.size

        if 'cell' in self.stabilizer_type(location):
            pauli = 'Z'
        else:
            pauli = 'X'

        x, y, z = location

        if 'cell' in self.stabilizer_type(location):
            # Cell coordinates consist of all permutations of {0,1,2} with
            # all possible signs in front of 1 and 2
            signs = np.array(list(itertools.product([-1, 1], [-1, 1])))
            permutations = np.vstack([signs * [1, 2], signs * [2, 1]])
            delta = np.vstack([np.insert(permutations, 0, 0, axis=1),
                               np.insert(permutations, 1, 0, axis=1),
                               np.insert(permutations, 2, 0, axis=1)]
                              ).tolist()

        elif self.stabilizer_type(location) == 'face-square':
            if x % 4 == z % 4:  # xz square
                delta = [(0, 0, -1), (0, 0, 1), (-1, 0, 0), (1, 0, 0)]
            elif y % 4 == z % 4:  # yz square
                delta = [(0, 0, -1), (0, 0, 1), (0, -1, 0), (0, 1, 0)]
            else:  # xy square
                delta = [(-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0)]

        elif self.stabilizer_type(location) == 'face-hex':
            if x % 4 == z % 4 and y % 4 == z % 4:
                # print("Hex 1")
                delta = [(-1, 0, 1), (1, 0, -1),
                         (0, -1, 1), (0, 1, -1),
                         (-1, 1, 0), (1, -1, 0)]
            elif x % 4 != z % 4 and y % 4 == z % 4:
                # print("Hex 2")
                delta = [(-1, 0, -1), (1, 0, 1),
                         (0, -1, 1), (0, 1, -1),
                         (-1, -1, 0), (1, 1, 0)]
            elif x % 4 == z % 4 and y % 4 != z % 4:
                # print("Hex 3")
                delta = [(-1, 0, 1), (1, 0, -1),
                         (0, -1, -1), (0, 1, 1),
                         (-1, -1, 0), (1, 1, 0)]
            else:
                # print("Hex 4")
                delta = [(-1, 0, -1), (1, 0, 1),
                         (0, -1, -1), (0, 1, 1),
                         (-1, 1, 0), (1, -1, 0)]

        operator: Operator = dict()
        for d in delta:
            qubit_location = (int(x + d[0]) % (4 * Lx),
                              int(y + d[1]) % (4 * Ly),
                              int(z + d[2]) % (4 * Lz))

            operator[qubit_location] = pauli

        return operator

    def qubit_axis(self, location):
        x, y, z = location

        return 'x'

    def get_logicals_z(self) -> List[Operator]:
        """The 3 logical Z operators (errors on all squares
        of a given layer)"""

        Lx, Ly, Lz = self.size

        logicals = []

        # Yellow-red membrane yz plane
        x = 0
        operator: Operator = dict()
        for y in range(2, 4*Ly, 4):
            for z in range(2, 4*Lz, 4):
                operator[(x, (y + 1) % (4 * Ly), z)] = 'Z'
                operator[(x, (y - 1) % (4 * Ly), z)] = 'Z'
                operator[(x, y, (z + 1) % (4 * Lz))] = 'Z'
                operator[(x, y, (z - 1) % (4 * Lz))] = 'Z'
        logicals.append(operator)

        # Green-blue membrane yz plane
        x = 2
        operator = dict()
        for y in range(0, 4*Ly, 4):
            for z in range(0, 4*Lz, 4):
                operator[(x, (y + 1) % (4 * Ly), z)] = 'Z'
                operator[(x, (y - 1) % (4 * Ly), z)] = 'Z'
                operator[(x, y, (z + 1) % (4 * Lz))] = 'Z'
                operator[(x, y, (z - 1) % (4 * Lz))] = 'Z'
        logicals.append(operator)

        # Red-green membrane yz plane
        x = 3
        operator = dict()
        for y in range(1, 4*Ly+2, 2):
            for z in range(1, 4*Lz+2, 2):
                cell_loc_1 = ((x-1) % (4*Lx), (y-1) % (4 * Ly), (z+1) % (4*Lz))
                cell_loc_2 = ((x-1) % (4*Lx), (y+1) % (4 * Ly), (z-1) % (4*Lz))
                if ((self.is_stabilizer(cell_loc_1) and
                        self.stabilizer_type(cell_loc_1) == 'cell-red') or
                        (self.is_stabilizer(cell_loc_2) and
                         self.stabilizer_type(cell_loc_2) == 'cell-red')):
                    for qubit_loc in self.get_stabilizer((x, y, z)).keys():
                        operator[qubit_loc] = 'Z'
        logicals.append(operator)

        # Yellow-red membrane xz plane
        y = 0
        operator = dict()
        for x in range(2, 4*Lx, 4):
            for z in range(2, 4*Lz, 4):
                operator[((x + 1) % (4 * Lx), y, z)] = 'Z'
                operator[((x - 1) % (4 * Lx), y, z)] = 'Z'
                operator[(x, y, (z + 1) % (4 * Lz))] = 'Z'
                operator[(x, y, (z - 1) % (4 * Lz))] = 'Z'
        logicals.append(operator)

        # Green-blue membrane xz plane
        y = 2
        operator = dict()
        for x in range(0, 4*Lx, 4):
            for z in range(0, 4*Lz, 4):
                operator[((x + 1) % (4*Lx), y, z)] = 'Z'
                operator[((x - 1) % (4*Lx), y, z)] = 'Z'
                operator[(x, y, (z + 1) % (4 * Lz))] = 'Z'
                operator[(x, y, (z - 1) % (4 * Lz))] = 'Z'
        logicals.append(operator)

        # Red and green membrane xz plane
        y = 3
        operator = dict()
        for x in range(1, 4*Lx+2, 2):
            for z in range(1, 4*Lz+2, 2):
                cell_loc_1 = ((x-1) % (4*Lx), (y-1) % (4 * Ly), (z+1) % (4*Lz))
                cell_loc_2 = ((x+1) % (4*Lx), (y-1) % (4 * Ly), (z-1) % (4*Lz))
                if ((self.is_stabilizer(cell_loc_1) and
                        self.stabilizer_type(cell_loc_1) == 'cell-red') or
                        (self.is_stabilizer(cell_loc_2) and
                         self.stabilizer_type(cell_loc_2) == 'cell-red')):
                    for qubit_loc in self.get_stabilizer((x, y, z)).keys():
                        operator[qubit_loc] = 'Z'
        logicals.append(operator)

        # Yellow-red membrane xy plane
        z = 0
        operator = dict()
        for x in range(2, 4*Lx, 4):
            for y in range(2, 4*Ly, 4):
                operator[((x + 1) % (4 * Lx), y, z)] = 'Z'
                operator[((x - 1) % (4 * Lx), y, z)] = 'Z'
                operator[(x, (y + 1) % (4 * Ly), z)] = 'Z'
                operator[(x, (y - 1) % (4 * Ly), z)] = 'Z'
        logicals.append(operator)

        # Green-blue membrane xy plane
        z = 2
        operator = dict()
        for x in range(0, 4*Lx, 4):
            for y in range(0, 4*Ly, 4):
                operator[((x + 1) % (4 * Lx), y, z)] = 'Z'
                operator[((x - 1) % (4 * Lx), y, z)] = 'Z'
                operator[(x, (y + 1) % (4 * Ly), z)] = 'Z'
                operator[(x, (y - 1) % (4 * Ly), z)] = 'Z'
        logicals.append(operator)

        # Red and green membrane xy plane
        z = 3
        operator = dict()
        for x in range(1, 4*Lx+2, 2):
            for y in range(1, 4*Ly+2, 2):
                cell_loc_1 = ((x-1) % (4*Lx), (y+1) % (4 * Ly), (z-1) % (4*Lz))
                cell_loc_2 = ((x+1) % (4*Lx), (y-1) % (4 * Ly), (z-1) % (4*Lz))
                if ((self.is_stabilizer(cell_loc_1) and
                        self.stabilizer_type(cell_loc_1) == 'cell-red') or
                        (self.is_stabilizer(cell_loc_2) and
                         self.stabilizer_type(cell_loc_2) == 'cell-red')):
                    for qubit_loc in self.get_stabilizer((x, y, z)).keys():
                        operator[qubit_loc] = 'Z'
        logicals.append(operator)

        return logicals

    def get_logicals_x(self) -> List[Operator]:
        """Get the 3 logical X operators."""
        Lx, Ly, Lz = self.size
        logicals = []

        operator: Operator = dict()
        for x in range(2, 4*Lx-2, 8):
            operator[(x-2, 6, 1)] = 'X'
            operator[(x-1, 6, 0)] = 'X'
            operator[(x+1, 6, 0)] = 'X'
            operator[(x+2, 6, 1)] = 'X'
        logicals.append(operator)

        operator = dict()
        for x in range(4, 4*Lx-2, 8):
            operator[(x-2, 0, 1)] = 'X'
            operator[(x-1, 0, 2)] = 'X'
            operator[(x+1, 0, 2)] = 'X'
            operator[(x+2, 0, 1)] = 'X'
        logicals.append(operator)

        operator = dict()
        for x in range(1, 4*Lx, 2):
            operator[(x, 2, 0)] = 'X'
        logicals.append(operator)

        operator = dict()
        for y in range(2, 4*Ly-2, 8):
            operator[(6, y-2, 1)] = 'X'
            operator[(6, y-1, 0)] = 'X'
            operator[(6, y+1, 0)] = 'X'
            operator[(6, y+2, 1)] = 'X'
        logicals.append(operator)

        operator = dict()
        for y in range(4, 4*Ly-2, 8):
            operator[(0, y-2, 1)] = 'X'
            operator[(0, y-1, 2)] = 'X'
            operator[(0, y+1, 2)] = 'X'
            operator[(0, y+2, 1)] = 'X'
        logicals.append(operator)

        operator = dict()
        for y in range(1, 4*Ly, 2):
            operator[(2, y, 0)] = 'X'
        logicals.append(operator)

        operator = dict()
        for z in range(2, 4*Lz-2, 8):
            operator[(6, 1, z-2)] = 'X'
            operator[(6, 0, z-1)] = 'X'
            operator[(6, 0, z+1)] = 'X'
            operator[(6, 1, z+2)] = 'X'
        logicals.append(operator)

        operator = dict()
        for z in range(4, 4*Lz-2, 8):
            operator[(0, 1, z-2)] = 'X'
            operator[(0, 2, z-1)] = 'X'
            operator[(0, 2, z+1)] = 'X'
            operator[(0, 1, z+2)] = 'X'
        logicals.append(operator)

        operator = dict()
        for z in range(1, 4*Lz, 2):
            operator[(2, 0, z)] = 'X'
        logicals.append(operator)

        return logicals

    def stabilizer_representation(
        self, location: Tuple, rotated_picture=False, json_file=None
    ) -> Dict:
        rep = super().stabilizer_representation(location, rotated_picture)

        x, y, z = location

        if self.stabilizer_type(location) == 'face-square':
            if x % 4 == z % 4:
                rep['params']['normal'] = [0, 1, 0]
            elif y % 4 == z % 4:
                rep['params']['normal'] = [1, 0, 0]
            else:
                rep['params']['normal'] = [0, 0, 1]
        elif self.stabilizer_type(location) == 'face-hex':
            # No idea where the sqrt(2)/2 comes from, but it seems necessary
            if x % 4 == z % 4 and y % 4 == z % 4:
                rep['params']['normal'] = [1, 1, np.sqrt(2)/2]
            elif x % 4 != z % 4 and y % 4 == z % 4:
                rep['params']['normal'] = [1, -1, -np.sqrt(2)/2]
            elif x % 4 == z % 4 and y % 4 != z % 4:
                rep['params']['normal'] = [1, -1, np.sqrt(2)/2]
            else:
                rep['params']['normal'] = [1, 1, -np.sqrt(2)/2]

        return rep



from abc import ABCMeta, abstractmethod
from typing import Optional, List
from panqec.codes import StabilizerCode
from panqec.error_models import BaseErrorModel
import numpy as np


class BaseDecoder(metaclass=ABCMeta):
    """Base class for decoders"""

    def __init__(
        self,
        code: StabilizerCode,
        error_model: BaseErrorModel,
        error_rate: float
    ):
        self.code = code
        self.error_model = error_model
        self.error_rate = error_rate

    @property
    @abstractmethod
    def allowed_codes(self) -> Optional[List[str]]:
        """List of codes concerned by the decoders (by id).
        Takes the value None if all codes are allowed.

        Example: ['Toric2DCode', 'Planar2DCode', 'RotatedPlanar2DCode']
        """

    @property
    @abstractmethod
    def label(self) -> str:
        """Label used in plots and result files
        E.g. 'Toric 2D Matching'
        """

    @property
    def id(self) -> str:
        """Class name"""
        return self.__class__.__name__

    @property
    @abstractmethod
    def params(self) -> dict:
        """List of class arguments (as a dictionary), that can be saved
        and reused to instantiate the same decoder.
        It should not include `code`, `error_model` and `error_rate`

        Example: `{'num_iterations': 10}`
        """

    @abstractmethod
    def decode(self, syndrome: np.ndarray, **kwargs) -> np.ndarray:
        """Given a code and a syndrome, returns a correction to apply
        to the qubits

        Parameters
        ----------
        syndrome: np.ndarray
            Syndrome as an array of size m, where m is the number of
            stabilizers. Each element contains 1 if the stabilizer is
            activated and 0 otherwise

        kwargs: dict
            Decoder-specific parameters (implemented by subclasses)

        Returns
        -------
        correction : np.ndarray
            Correction as an array of size 2n (with n the number of qubits)
            in the binary symplectic format.
        """



import pytest
from panqec.codes import HollowPlanar3DCode
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestHollowPlanar3DCode(StabilizerCodeTest):

    @pytest.fixture(params=[(2, 2, 2), (3, 3, 3), (2, 3, 4)])
    def code(self, request):
        return HollowPlanar3DCode(*request.param)



from typing import Tuple, Dict, List
from panqec.codes import StabilizerCode
import numpy as np

Operator = Dict[Tuple, str]  # Location to pauli ('X', 'Y' or 'Z')
Coordinates = List[Tuple]  # List of locations


class Planar2DCode(StabilizerCode):
    """2D surface code with open boundary conditions.
    In this variant of the 2D surface code,
    there are two types of boundaries: rough and smooth.

    As in the Kitaev toric code,
    the qubits live on edges of the lattice,
    while stabilizers are defined on faces and vertices.
    As such, on rough boundaries the edge-most qubits are on
    edges orthogonal to the boundary forming a 'rough' look.
    On smooth boundaries the edge-most qubits are on
    edges parallel to the boundary forming a 'smooth' look.
    The boundary conditions and coordinate system used are
    shown below.

    .. image:: planar_2d_code.svg
        :scale: 200 %
        :align: center

    Parameters
    ----------
    L_x : int
        The size in the x direction.
    L_y : Optional[int]
        The size in the y direction.
        If it is not given, it is assumed to be a square lattice with Lx=Ly.
    """
    dimension = 2
    deformation_names = ['XZZX', 'XY']

    @property
    def label(self) -> str:
        return 'Planar {}x{}'.format(*self.size)

    def get_qubit_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly = self.size

        # Qubits along e_x
        for x in range(1, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                coordinates.append((x, y))

        # Qubits along e_y
        for x in range(2, 2*Lx, 2):
            for y in range(1, 2*Ly-1, 2):
                coordinates.append((x, y))

        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly = self.size

        # Vertices
        for x in range(2, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                coordinates.append((x, y))

        # Faces
        for x in range(1, 2*Lx, 2):
            for y in range(1, 2*Ly-1, 2):
                coordinates.append((x, y))

        return coordinates

    def stabilizer_type(self, location: Tuple) -> str:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        x, y = location
        if x % 2 == 0:
            return 'vertex'
        else:
            return 'face'

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        if self.stabilizer_type(location) == 'vertex':
            pauli = 'Z'
        else:
            pauli = 'X'

        delta: List[Tuple] = [(-1, 0), (1, 0), (0, -1), (0, 1)]

        operator: Operator = dict()
        for d in delta:
            qubit_location = tuple(np.add(location, d))

            if self.is_qubit(qubit_location):
                operator[qubit_location] = pauli

        return operator

    def qubit_axis(self, location: Tuple) -> str:
        x, y = location

        if (x % 2 == 1) and (y % 2 == 0):
            axis = 'x'
        elif (x % 2 == 0) and (y % 2 == 1):
            axis = 'y'
        else:
            raise ValueError(f'Location {location} does not correspond'
                             'to a qubit')

        return axis

    def get_logicals_x(self) -> List[Operator]:
        Lx, Ly = self.size
        logicals: List[Operator] = []

        # X operators along x edges in x direction.
        operator: Operator = dict()
        for x in range(1, 2*Lx, 2):
            operator[(x, 0)] = 'X'
        logicals.append(operator)

        return logicals

    def get_logicals_z(self) -> List[Operator]:
        Lx, Ly = self.size
        logicals: List[Operator] = []

        # X operators along x edges in x direction.
        operator: Operator = dict()
        for y in range(0, 2*Ly, 2):
            operator[(1, y)] = 'Z'
        logicals.append(operator)

        return logicals

    def get_deformation(
        self, location: Tuple,
        deformation_name: str,
        deformation_axis: str = 'y',
        **kwargs
    ) -> Dict:

        if deformation_axis not in ['x', 'y']:
            raise ValueError(f"{deformation_axis} is not a valid "
                             "deformation axis")

        if deformation_name == 'XZZX':
            undeformed_dict = {'X': 'X', 'Y': 'Y', 'Z': 'Z'}
            deformed_dict = {'X': 'Z', 'Y': 'Y', 'Z': 'X'}

            if self.qubit_axis(location) == deformation_axis:
                deformation = deformed_dict
            else:
                deformation = undeformed_dict

        elif deformation_name == 'XY':
            deformation = {'X': 'X', 'Y': 'Z', 'Z': 'Y'}

        else:
            raise ValueError(f"The deformation {deformation_name}"
                             "does not exist")

        return deformation



"""
Tests for custom implementation of pauli bit strings.

:Author:
    Eric Huang
"""
import pytest
import numpy as np
from panqec.codes import Toric2DCode
from panqec.bpauli import (
    pauli_string_to_bvector, bvector_to_pauli_string,
    bs_prod, get_effective_error, bvector_to_int,
    bvectors_to_ints, ints_to_bvectors, apply_deformation
)
from panqec.bsparse import from_array, is_sparse, vstack


def test_pauli_string_to_bvector():
    assert np.all(
        pauli_string_to_bvector('IXYZ') == [0, 1, 1, 0, 0, 0, 1, 1]
    ), 'bsf of IXYZ should be 01100011'


def test_bvector_to_pauli_string():
    assert (
        bvector_to_pauli_string(np.array([0, 1, 1, 0, 0, 0, 1, 1])) == 'IXYZ'
    ), 'Pauli string of 01100011 should be IXYZ'


def test_pauli_string_bvector_inverse():
    pstring = 'IXYZ'
    bvector = pauli_string_to_bvector(pstring)
    new_pstring = bvector_to_pauli_string(bvector)
    assert pstring == new_pstring, (
        'Converting Pauli string to bsf and back should '
        'give original Pauli string'
    )


class Testbs_prod:

    def test_bs_prod_singles(self):
        III = np.array([0, 0, 0, 0, 0, 0])
        XXX = np.array([1, 1, 1, 0, 0, 0])
        ZZZ = np.array([0, 0, 0, 1, 1, 1])
        assert bs_prod(XXX, ZZZ) == 1, 'XXX should anticommute with ZZZ'
        assert bs_prod(III, XXX) == 0, 'III should commute with XXX'
        assert bs_prod(III, XXX) == 0, 'III should commute with XXX'

    def test_bs_prod_many_to_one(self):
        stabilizers = np.array([
            pauli_string_to_bvector('XXI'),
            pauli_string_to_bvector('IXX'),
        ])
        error = pauli_string_to_bvector('IZI')
        syndrome = bs_prod(stabilizers, error)
        assert syndrome.shape == (2,), 'syndrome should have shape 2'
        assert np.all(syndrome == [1, 1]), 'IZI should anticommute with both'

        # Changing the order shouldn't matter.
        syndrome = bs_prod(error, stabilizers)
        assert syndrome.shape == (2,)
        assert np.all(syndrome == [1, 1])

    def test_bs_prod_one_to_many(self):
        XYZ = pauli_string_to_bvector('XYZ')
        IXY = pauli_string_to_bvector('IXY')
        ZZI = pauli_string_to_bvector('ZZI')
        assert bs_prod(XYZ, IXY) == 0
        assert bs_prod(IXY, ZZI) == 1
        assert np.all(bs_prod(XYZ, [IXY, ZZI]) == [[0, 0]])
        assert np.all(bs_prod([XYZ, IXY], [ZZI, IXY]) == [[0, 0], [1, 0]])

    def test_raise_error_if_not_even_length(self):
        with pytest.raises(ValueError):
            bs_prod([0, 0, 1, 0, 1], [0, 1, 0, 1, 0])

        with pytest.raises(ValueError):
            bs_prod([0, 0, 0], [0, 1])

        with pytest.raises(ValueError):
            bs_prod([0, 0, 0, 0], [0, 1, 0])

    def test_raise_error_if_unequal_shapes(self):
        with pytest.raises(ValueError):
            bs_prod([0, 0, 0, 1], [1, 0, 1, 1, 0, 1])


class Testbs_prodSparse:

    def test_bs_prod_singles(self):
        III = from_array(np.array([0, 0, 0, 0, 0, 0]))
        XXX = from_array(np.array([1, 1, 1, 0, 0, 0]))
        ZZZ = from_array(np.array([0, 0, 0, 1, 1, 1]))
        assert is_sparse(III)
        assert bs_prod(XXX, ZZZ) == 1, 'XXX should anticommute with ZZZ'
        assert bs_prod(III, XXX) == 0, 'III should commute with XXX'
        assert bs_prod(III, XXX) == 0, 'III should commute with XXX'

    def test_bs_prod_many_to_one(self):
        stabilizers = from_array(np.array([
            pauli_string_to_bvector('XXI'),
            pauli_string_to_bvector('IXX'),
        ]))
        assert stabilizers.shape == (2, 6)
        error = from_array(pauli_string_to_bvector('IZI'))
        syndrome = bs_prod(stabilizers, error)
        assert syndrome.shape == (2,), 'syndrome should have shape 2'
        assert np.all(syndrome == [1, 1]), 'IZI should anticommute with both'

        # Changing the order shouldn't matter.
        syndrome = bs_prod(error, stabilizers)
        assert syndrome.shape == (2,)
        assert np.all(syndrome == [1, 1])

    def test_bs_prod_one_to_many(self):
        XYZ = from_array(pauli_string_to_bvector('XYZ'))
        IXY = from_array(pauli_string_to_bvector('IXY'))
        ZZI = from_array(pauli_string_to_bvector('ZZI'))
        assert bs_prod(XYZ, IXY) == 0
        assert bs_prod(IXY, ZZI) == 1
        assert np.all(bs_prod(XYZ, vstack([IXY, ZZI])) == [[0, 0]])
        assert np.all(
            bs_prod(vstack([XYZ, IXY]), vstack([ZZI, IXY]))
            == [
                [0, 0],
                [1, 0]
            ]
        )


def test_effective_error_wrong_size_raises_exception():
    X_logicals = np.array([1, 0, 1], dtype=np.uint)
    Z_logicals = np.array([1, 0, 1, 0], dtype=np.uint)
    total_error = np.array([0, 1], dtype=np.uint)
    with pytest.raises(ValueError):
        get_effective_error(total_error, X_logicals, Z_logicals)


def test_get_effective_errror_single():
    X_logicals = pauli_string_to_bvector('XXXXX')
    Z_logicals = pauli_string_to_bvector('ZZZZZ')
    total_error = pauli_string_to_bvector('YYYYY')
    effective_error = get_effective_error(total_error, X_logicals, Z_logicals)
    assert np.all(effective_error.shape == (2, ))
    assert bvector_to_pauli_string(effective_error) == 'Y'


def test_get_effective_error_many():
    X_logicals = pauli_string_to_bvector('XXXXX')
    Z_logicals = pauli_string_to_bvector('ZZZZZ')
    total_error = np.array([
        pauli_string_to_bvector('YYYYY'),
        pauli_string_to_bvector('IIIII'),
        pauli_string_to_bvector('XXXZZ'),
    ])
    effective_error = get_effective_error(total_error, X_logicals, Z_logicals)
    assert np.all(effective_error.shape == (3, 2)), (
        'Effective errors should have shape (3, 2), '
        f'instead got {effective_error.shape}'
    )
    assert np.all(effective_error == np.array([
        pauli_string_to_bvector('Y'),
        pauli_string_to_bvector('I'),
        pauli_string_to_bvector('X')
    ])), 'Effective errors should be bsf for Y, I, Z'


def test_bvector_to_int():
    assert bvector_to_int(pauli_string_to_bvector('IIIII')) == 0
    assert bvector_to_int(pauli_string_to_bvector('I')) == 0
    assert bvector_to_int(pauli_string_to_bvector('X')) == 2
    assert bvector_to_int(pauli_string_to_bvector('Y')) == 3
    assert bvector_to_int(pauli_string_to_bvector('Z')) == 1


def test_bvectors_to_ints():
    assert bvectors_to_ints(list(map(
        pauli_string_to_bvector,
        ['III', 'XYZ', 'IIZ']
    ))) == [0, 51, 1]


def test_ints_to_bvectors():
    assert np.all(
        np.array(ints_to_bvectors([0, 1, 2], 3))
        == np.array(list(map(
            pauli_string_to_bvector, ['III', 'IIZ', 'IZI']
        )))
    )


def test_get_effective_error_toric_code_logicals():
    code = Toric2DCode(3, 5)
    logical_operators = {
        'X1': code.logicals_x[0],
        'X2': code.logicals_x[1],
        'Z1': code.logicals_z[0],
        'Z2': code.logicals_z[1],
    }
    expected_effective_errors = {
        'X1': [1, 0, 0, 0],
        'X2': [0, 1, 0, 0],
        'Z1': [0, 0, 1, 0],
        'Z2': [0, 0, 0, 1],
    }
    for logical in ['X1', 'X2', 'Z1', 'Z2']:
        effective_error = get_effective_error(
            logical_operators[logical], code.logicals_x, code.logicals_z
        )
        assert np.all(effective_error == expected_effective_errors[logical]), (
            f'Logical operator {logical} should have effective error '
            f'{expected_effective_errors[logical]}, '
            f'instead got {effective_error.tolist()}'
        )


def test_get_effective_error_toric_code_logicals_many():
    code = Toric2DCode(3, 5)
    logical_operators = np.array([
        code.logicals_x[0],
        code.logicals_x[1],
        code.logicals_z[0],
        code.logicals_z[1],
    ])
    expected_effective_errors = np.array([
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ])
    effective_errors = get_effective_error(
        logical_operators, code.logicals_x, code.logicals_z
    )
    assert np.all(effective_errors == expected_effective_errors), (
        f'Logical operators {logical_operators} should have effective errors '
        f'{expected_effective_errors}, '
        f'instead got {effective_errors}'
    )


class TestApplyDeformation:

    @pytest.mark.parametrize('deformation_index,bsf,deformed', [
        (
            [False, False, False],
            [1, 1, 0, 0, 1, 1],
            [1, 1, 0, 0, 1, 1],
        ),
        (
            [False, False, True],
            [1, 1, 0, 0, 1, 1],
            [1, 1, 1, 0, 1, 0],
        ),
        (
            [False, True, True],
            [1, 1, 0, 0, 1, 1],
            [1, 1, 1, 0, 1, 0],
        ),
        (
            [True, True, True],
            [1, 1, 0, 0, 1, 1],
            [0, 1, 1, 1, 1, 0],
        ),
        (
            [True, False, False],
            [
                [1, 1, 0, 0, 1, 1],
                [1, 0, 1, 1, 1, 0],
                [0, 1, 1, 1, 0, 1],
            ],
            [
                [0, 1, 0, 1, 1, 1],
                [1, 0, 1, 1, 1, 0],
                [1, 1, 1, 0, 0, 1],
            ],
        ),
    ])
    def test_apply_deformation_on_easy_examples(
        self, deformation_index, bsf, deformed
    ):
        bsf = np.array(bsf, dtype=np.uint)
        deformed = np.array(deformed, dtype=np.uint)
        assert np.all(deformed == apply_deformation(deformation_index, bsf))

    def test_apply_deformation_preserves_commutators(self):
        operators = np.array([
            pauli_string_to_bvector(pauli_string)
            for pauli_string in ['IXYZ', 'XXXX', 'ZXYI', 'IIZZ']
        ], dtype=np.uint)
        commutators = bs_prod(operators, operators)
        deformation_index = [True, False, True, False]
        deformed_commutators = bs_prod(
            apply_deformation(deformation_index, operators),
            apply_deformation(deformation_index, operators)
        )
        assert np.all(commutators == deformed_commutators)

    def test_apply_deformation_is_involution(self):
        original = np.array([
            pauli_string_to_bvector(pauli_string)
            for pauli_string in ['IXYZ', 'XXXX', 'ZXYI', 'IIZZ']
        ], dtype=np.uint)
        deformation_index = [True, False, True, False]
        deformed_once = apply_deformation(deformation_index, original)
        deformed_twice = apply_deformation(deformation_index, deformed_once)
        assert np.all(deformed_twice == original)

    def test_value_error_raised_single_bsf(self):
        bsf = pauli_string_to_bvector('XYZ')
        deformation_index = [True, False]
        with pytest.raises(ValueError):
            apply_deformation(deformation_index, bsf)

    def test_value_error_raised_many_bsf(self):
        bsf = np.array([
            pauli_string_to_bvector(pauli_string)
            for pauli_string in ['IXYZ', 'XXXX', 'ZXYI', 'IIZZ']
        ], dtype=np.uint)
        deformation_index = [True, False, True]
        with pytest.raises(ValueError):
            apply_deformation(deformation_index, bsf)



"""API for running simulations."""

import os
import json
from json import JSONDecodeError
import datetime
import itertools
from typing import List, Dict, Callable, Union, Any, Optional, Tuple, Iterable
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from panqec.codes import StabilizerCode
from panqec.decoders import BaseDecoder
from panqec.error_models import BaseErrorModel
from panqec.config import (
    CODES, ERROR_MODELS, DECODERS
)
from panqec.utils import identity, load_json, save_json
from . import (
    BaseSimulation, DirectSimulation, SplittingSimulation
)
from panqec.analysis import Analysis


def run_file(
    input_file: str,
    output_file: str,
    n_trials: int,
    progress: Callable = identity,
    log_file: Optional[str] = None,
    verbose: bool = True,
):
    """Run an input json file.

    Parameters
    ----------
    input_file : str
        Path to the input json file.
    output_file: str
        Path to the json file that will contain the results.
    n_trials : int
        The number of MCMC trials to do.
    progress : Callable
        Callable function
    verbose: bool,
        Verbosity of the output
    Returns
    -------
    None
    """
    print(f"Run file {input_file}")

    batch_sim = read_input_json(input_file, output_file, log_file=log_file)

    if verbose:
        print(f'running {len(batch_sim._simulations)} simulations:')
        for simulation in batch_sim._simulations:
            code = simulation.code.label
            noise = simulation.error_model.label
            if isinstance(simulation, SplittingSimulation):
                decoder = simulation.decoders[0].label
                error_rates = simulation.error_rates
                print(f'{code}, {noise}, {decoder}, {error_rates}')
            elif isinstance(simulation, DirectSimulation):
                decoder = simulation.decoder.label
                error_rate = simulation.error_rate
                print(f'{code}, {noise}, {decoder}, {error_rate}')
    batch_sim.run(n_trials, progress=progress)


class BatchSimulation():
    """Holds and controls many simulations.

    Parameters
    ----------
    label : str
        The label of the files.
    output_file : str
        Path to the file (.json or .json.gz) that will store
        the simulation results
    label: str
        Label of the batch simulation
    on_update : Callable
        Function that gets called on every update.
    update_frequency : int
        Frequency at which to update results.
    save_frequency : int
        Frequency at which to write results to file on disk.
    method: str
        The method can be either "direct" or "splitting".
        The direct method samples independent errors at each iteration,
        while the splitting method (by Bravyi & Vargo) uses MCMC to
        determine the next error to sample.
        Ref: arXiv:1308.6270
    """

    _simulations: List[BaseSimulation]
    update_frequency: int
    save_frequency: int
    _output_file: str

    def __init__(
        self,
        output_file: str,
        label='unlabelled',
        on_update: Callable = identity,
        update_frequency: int = 5,
        save_frequency: int = 1,
        method: str = "direct",
        log_file: Optional[str] = None,
        verbose: bool = True,
    ):
        self._simulations = []
        self.code: Dict = {}
        self.decoder: Dict = {}
        self.update_frequency = update_frequency
        self.save_frequency = save_frequency
        self.label = label
        self.method = method
        self.verbose = verbose
        self._output_file = output_file
        self._log_file = log_file

    def __getitem__(self, *args):
        return self._simulations.__getitem__(*args)

    def __iter__(self):
        return self._simulations.__iter__()

    def __next__(self):
        return self._simulations.__next__()

    def __len__(self):
        return self._simulations.__len__()

    def append(self, simulation: BaseSimulation):
        """Append a simulation to the current BatchSimulation.

        Parameters
        ----------
        simulation: BaseSimulation
            The simulation to append.
        """
        self._simulations.append(simulation)

    def load_results(self):
        """Load results from disk."""
        for simulation in self._simulations:
            simulation.load_results(self._output_file)

    def on_update(self, n_trials: int):
        """Function that gets called on every update.
        It uses the total number of runs, `n_trials`, to estimate
        the remaining time
        """
        pass

    def estimate_remaining_time(self, n_trials: int):
        """Estimate remaining time given target n_trials.

        Parameters
        ----------
        n_trials : int
            The target total number of trials.
        """
        return sum(
            (n_trials - sim.n_results)*sim.wall_time/sim.n_results
            for sim in self._simulations
            if sim.n_results != 0
        )

    @property
    def wall_time(self):
        """Total time run so far."""
        return sum(sim.wall_time for sim in self._simulations)

    def run(self, n_trials, progress: Callable = identity):
        """Perform the running.

        Parameters
        ----------
        n_trials : int
            Number of trials to run.
        progress : Callable
            The progress bar, such as tqdm.
        """
        try:
            self._run(n_trials, progress=progress)
        except KeyboardInterrupt:
            print('Simulation paused')

    def _run(self, n_trials, progress: Callable = identity):
        self.load_results()

        # Use the maximum remaining trials to overestimate how much is
        # remaining for purposes of reporting progress to give a conservative
        # estimate of what is left without disappointing the user.
        min_current_trial = min([
            simulation.n_results for simulation in self._simulations
        ])

        for i_trial in progress(list(range(min_current_trial, n_trials))):
            for simulation in self._simulations:
                if simulation.n_results < n_trials:
                    simulation.run(1)
            if i_trial > 0:
                if i_trial % self.update_frequency == 0:
                    self.on_update(n_trials)
                if i_trial % self.save_frequency == 0:
                    self.save_results()
            if i_trial == n_trials - 1:
                self.on_update(n_trials)
                self.save_results()

            self._log_progress(i_trial, n_trials)

        # for simulation in self._simulations:
        #     if self.verbose:
        #         print(f"\nPost-processing {simulation.label}")
        #     simulation.postprocess()

    def _save_results(self):
        self._update_file(
            self.get_results_to_save()
        )

    def save_file(self):
        """Do a complete save of the file."""
        combined_data = []
        for simulation in self._simulations:
            combined_data.append(simulation.get_results_to_save())

        # Create output directory if it does not exist yet
        output_dir = os.path.dirname(self._output_file)
        if output_dir == '':
            output_dir = os.getcwd()

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        save_json(combined_data, self._output_file)

    def _log_progress(self, i_trial, n_trials):
        if self._log_file is not None:
            with open(self._log_file, "w") as f:
                f.write(f"{i_trial+1}/{n_trials}")

    def _update_file(self, new_data: list) -> None:
        """Update only the results to one file."""

        # First time file does not exist, so write it to start.
        if not os.path.isfile(self._output_file):
            self.save_file()

        # Write the updated list to the .json or .json.gz file.
        save_json(new_data, self._output_file)

    def load_file(self):
        pass

    def save_results(self):
        try:
            self._save_results()

        # Do not give up saving results during keyboard interrupt.
        except KeyboardInterrupt:
            print('Simulation paused. Saving results. Do not interrupt again')
            self._save_results()
            print('Results saved')
            raise KeyboardInterrupt('Simulation paused')

    def get_results(self):
        results = []
        for simulation in self._simulations:
            simulation_data = simulation.get_results()
            results.append(simulation_data)
        return results

    def get_results_to_save(self):
        results = []
        for simulation in self._simulations:
            simulation_data = simulation.get_results_to_save()
            results.append(simulation_data)
        return results

    def get_results_df(self):
        batch_results = self.get_results()
        # print(
        #     'wall_time =',
        #     str(datetime.timedelta(seconds=batch_sim.wall_time))
        # )
        # print('n_trials = ', min(sim.n_results for sim in batch_sim))
        for sim, batch_result in zip(self, batch_results):
            n_logicals = batch_result['k']

            # Small fix for the current situation. TO REMOVE in later versions
            if n_logicals == -1:
                n_logicals = 1

            batch_result['noise_direction'] = sim.error_model.direction

            if self.method == 'direct':
                if len(sim.results['effective_error']) > 0:
                    batch_result['p_x'] = np.array(
                        sim.results['effective_error']
                    )[:, :n_logicals].any(axis=1).mean()
                    batch_result['p_x_se'] = np.sqrt(
                        batch_result['p_x']*(1 - batch_result['p_x'])
                        / (sim.n_results + 1)
                    )
                    batch_result['p_z'] = np.array(
                        sim.results['effective_error']
                    )[:, n_logicals:].any(axis=1).mean()
                    batch_result['p_z_se'] = np.sqrt(
                        batch_result['p_z']*(1 - batch_result['p_z'])
                        / (sim.n_results + 1)
                    )
                else:
                    batch_result['p_x'] = np.nan
                    batch_result['p_x_se'] = np.nan
                    batch_result['p_z'] = np.nan
                    batch_result['p_z_se'] = np.nan

        results = batch_results

        results_df = pd.DataFrame(results)

        if self.method == 'splitting':
            results_df = results_df.explode(['error_rates', 'p_est', 'p_se'])

        return results_df

    def activate_live_update(self):
        self.on_update = self._update_plot

    def _update_plot(self, n_trials: int):
        import IPython

        plt.clf()

        remaining_time = self.estimate_remaining_time(n_trials)

        analysis = Analysis(self._output_file, verbose=False)
        analysis.plot_thresholds(
            include_threshold_estimate=False,
            include_main_title=False,
            include_sector_title=False
        )
        plt.title(
            f'Time remaining '
            f'{datetime.timedelta(seconds=int(remaining_time))}'
        )

        IPython.display.clear_output(wait=True)
        IPython.display.display(plt.gcf())


def _parse_parameters_range(parameters):
    parameters_range = [{}]
    if len(parameters) > 0:
        if isinstance(parameters, list):
            parameters_range = parameters
        elif isinstance(parameters, dict):
            parameters_range = [parameters]
        else:
            parameters_range = [parameters]
    return parameters_range


def _parse_all_ranges(data: dict) -> Tuple[list, list, list, list]:
    if 'parameters' in data['code']:
        params_range = _parse_parameters_range(data['code']['parameters'])
        code_range = []
        for params in params_range:
            code_range.append(data['code'].copy())
            code_range[-1]['parameters'] = params

    noise_range: List[Dict] = [{}]
    if 'parameters' in data['error_model']:
        params_range = _parse_parameters_range(
            data['error_model']['parameters']
        )
        noise_range = []
        for params in params_range:
            noise_range.append(data['error_model'].copy())
            noise_range[-1]['parameters'] = params

    decoder_range: List[Dict] = [{}]
    parameters = []
    if 'parameters' in data['decoder']:
        parameters = data['decoder']['parameters']

    params_range = _parse_parameters_range(parameters)
    decoder_range = []
    for params in params_range:
        decoder_range.append(data['decoder'].copy())
        decoder_range[-1]['parameters'] = params

    error_rate_range = _parse_parameters_range(data['error_rate'])

    return code_range, noise_range, decoder_range, error_rate_range


def expand_input_ranges(data: dict) -> List[Dict]:
    runs: List[Dict] = []

    (
        code_range, error_model_range, decoder_range, error_rate_range
    ) = _parse_all_ranges(data)

    for (
        error_model_param, decoder_param, error_rate, code_param
    ) in itertools.product(
        error_model_range, decoder_range, error_rate_range, code_range
    ):
        run = {
            k: v for k, v in data.items()
            if k not in ['code', 'error_model', 'decoder', 'error_rate']
        }
        for key in ['code', 'error_model', 'decoder']:
            run[key] = {
                k: v for k, v in data[key].items() if k != 'parameters'
            }
        run['code']['parameters'] = code_param['parameters']
        run['error_model']['parameters'] = error_model_param['parameters']
        run['decoder']['parameters'] = decoder_param['parameters']
        run['error_rate'] = error_rate

        runs.append(run)

    return runs


def _parse_code_dict(code_dict: Dict[str, Any]) -> StabilizerCode:
    code_name = code_dict['name']
    code_params: Union[list, dict] = []
    if 'parameters' in code_dict:
        code_params = code_dict['parameters']
    code_class = CODES[code_name]
    if isinstance(code_params, dict):
        code = code_class(**code_params)  # type: ignore
    else:
        code = code_class(*code_params)  # type: ignore
    return code


def _parse_error_model_dict(noise_dict: Dict[str, Any]) -> BaseErrorModel:
    error_model_name = noise_dict['name']
    error_model_params: Union[list, dict] = []
    if 'parameters' in noise_dict:
        error_model_params = noise_dict['parameters']
    error_model_class = ERROR_MODELS[error_model_name]
    if isinstance(error_model_params, dict):
        error_model = error_model_class(**error_model_params)
    else:
        error_model = error_model_class(*error_model_params)
    return error_model


def _parse_decoder_dict(
    decoder_dict: Dict[str, Any],
    code: StabilizerCode,
    error_model: BaseErrorModel,
    error_rate: float
) -> BaseDecoder:
    decoder_name = decoder_dict['name']
    decoder_class = DECODERS[decoder_name]
    decoder_params: dict = {}
    if 'parameters' in decoder_dict:
        decoder_params = decoder_dict['parameters']
    else:
        decoder_params = {}

    decoder_params['code'] = code
    decoder_params['error_model'] = error_model
    decoder_params['error_rate'] = error_rate

    decoder = decoder_class(**decoder_params)
    return decoder


def read_input_json(
    input_file: str,
    output_file: str,
    log_file: Optional[str] = None
) -> BatchSimulation:
    """Read json input file or .json.gz file."""
    try:
        data = load_json(input_file)
    except JSONDecodeError as err:
        print(f'Error reading input file {input_file}')
        raise err

    return read_input_dict(data, output_file, log_file=log_file)


def get_runs(data: dict) -> List[dict]:
    """Get expanded runs from input dictionary."""

    runs = []
    if 'runs' in data:
        runs = data['runs']
    if 'ranges' in data:
        runs += expand_input_ranges(data['ranges'])

    return runs


def count_runs(file_path: str) -> Optional[int]:
    """Count the number of noise parameters in an input file.

    Return None if no noise parameters range given.
    """
    n_runs = None
    with open(file_path) as f:
        data = json.load(f)
    all_runs = get_runs(data)
    n_runs = len(all_runs)
    return n_runs


def get_simulations(data: dict, verbose: bool = True) -> List[BaseSimulation]:
    simulations: List[BaseSimulation] = []

    method = 'direct'

    method_params = {}

    # The case if the ranges attribute is a list of dicts.
    # This allows everything to be jammed into one input file.
    if 'ranges' in data and isinstance(data['ranges'], list):
        for sub_ranges in data['ranges']:
            sub_data = dict(data)
            sub_data['ranges'] = sub_ranges
            simulations += get_simulations(sub_data)
        return simulations

    if 'ranges' in data:
        (
            code_range, noise_range, decoder_range, error_rates
        ) = _parse_all_ranges(data['ranges'])

        codes = [_parse_code_dict(code_dict) for code_dict in code_range]
        error_models = [_parse_error_model_dict(noise_dict)
                        for noise_dict in noise_range]
        instances: Iterable[Tuple] = itertools.product(
            codes, error_models, decoder_range, error_rates
        )

        if 'method' in data['ranges']:
            method = data['ranges']['method']['name']
            method_params = data['ranges']['method']['parameters']

    elif 'runs' in data:
        print("Run", data['runs'])
        codes = [_parse_code_dict(run['code']) for run in data['runs']]
        decoder_range = [run['decoder'] for run in data['runs']]
        error_models = [_parse_error_model_dict(run['error_model'])
                        for run in data['runs']]
        error_rates = [run['error_rate'] for run in data['runs']]
        instances = zip(codes, error_models, decoder_range, error_rates)

    else:
        raise ValueError("Invalid data format: does not have 'runs'\
                         or 'ranges' key")

    if method == 'direct':
        for code, error_model, decoder_dict, error_rate in instances:
            decoder = _parse_decoder_dict(decoder_dict, code, error_model,
                                          error_rate)

            simulations.append(DirectSimulation(code, error_model, decoder,
                                                error_rate, verbose=verbose,
                                                **method_params))

    if method == 'splitting':
        for code, error_model, decoder_dict in instances:
            decoders = [_parse_decoder_dict(decoder_dict, code, error_model, p)
                        for p in error_rates]

            simulations.append(SplittingSimulation(
                code, error_model, decoders, error_rates,
                verbose=verbose, **method_params
            ))

    return simulations


def read_input_dict(
    data: dict,
    output_file: str,
    verbose: bool = True,
    *args, **kwargs
) -> BatchSimulation:
    """Return BatchSimulation from input dict.

    Parameters
    ----------
    data : dict
        Data that has been parsed from an input json file.
    Returns
    -------
    batch_simulation : BatchSimulation
        BatchSimulation object populated with the simulations specified in the
        input data.
    """
    label = 'unlabelled'
    method = 'direct'

    if 'ranges' in data:
        # If many ranges are given label is 'combined', but if there is
        # only one unique label, then we can just use that label.
        if isinstance(data['ranges'], list):
            label = 'combined'
            labels = []
            for subdata in data['ranges']:
                if 'ranges' in subdata:
                    if 'label' in subdata['ranges']:
                        labels.append(subdata['ranges']['label'])
            if labels:
                unique_labels = list(set(labels))
                if len(unique_labels) == 1:
                    label = unique_labels[0]
        elif 'label' in data['ranges']:
            label = data['ranges']['label']

        if 'method' in data['ranges']:
            method = data['ranges']['method']['name']

    kwargs['label'] = label
    kwargs['method'] = method
    kwargs['verbose'] = verbose

    print("Start batch simulation instance")
    batch_sim = BatchSimulation(output_file, *args, **kwargs)
    assert len(batch_sim._simulations) == 0

    simulations = get_simulations(data, verbose=verbose)

    for sim in simulations:
        batch_sim.append(sim)

    return batch_sim



"""
Routines for input and output.

This includes routines to produce, merge and analyse data produced by
simulations.

:Author:
    Eric Huang
"""
import numpy as np
import datetime
import os
import json
from .bpauli import bvectors_to_ints
from .utils import sizeof_fmt


def serialize_results(
    i_trial: int, n_trials: int,
    L_list: np.ndarray, p_list: np.ndarray, L_repeats: np.ndarray,
    start_time: datetime.datetime, time_elapsed: datetime.timedelta,
    time_remaining: datetime.timedelta, eta: datetime.datetime,
    p_est: np.ndarray, p_se: np.ndarray,
    n_fail: np.ndarray, n_try: np.ndarray,
    effective_errors: list
) -> dict:
    """Convert results to dict."""
    return {
        'parameters': {
            'i_trial': i_trial,
            'n_trials': n_trials,
            'L_list': L_list.tolist(),
            'p_list': p_list.tolist(),
            'L_repeats': L_repeats.tolist(),
            'n_list': [int(3*L**3) for L in L_list],
        },
        'time': {
            'start_time': str(start_time),
            'time_elapsed': str(time_elapsed),
            'time_remaining': str(time_remaining),
            'eta': str(eta),
        },
        'statistics': {
            'p_est': p_est.tolist(),
            'p_se': p_se.tolist(),
        },
        'results': {
            'n_fail': n_fail.tolist(),
            'n_try': n_try.tolist(),
            'effective_errors': [
                [
                    bvectors_to_ints(effective_errors[i_L][i_p])
                    for i_p in range(len(p_list))
                ]
                for i_L in range(len(L_list))
            ],
        },
    }


def dump_results(export_json: str, results_dict: dict, verbose: bool = True):
    """Save results dict to json file."""
    with open(export_json, 'w') as f:
        json.dump(results_dict, f)
    if verbose:
        print(
            f'Results written to {export_json} '
            f'({sizeof_fmt(os.path.getsize(export_json))})'
        )



from typing import Tuple, Dict, List
import numpy as np
from panqec.codes import StabilizerCode

Operator = Dict[Tuple, str]  # Location to pauli ('X','Y','Z')
Coordinates = List[Tuple]  # List of locations


class RotatedPlanar2DCode(StabilizerCode):
    """Rotated 2D surface code with open boundaries.
    These use roughly half the number of qubits but only encode 1 logical
    qubit.

    Parameters
    ----------
    L_x : int
        Number of qubits in the x direction.
    L_y : Optional[int]
        Number of qubits in the y direction.
        Assumed square if not given.

    Notes
    -----
    One stabilizer of each type is shown in the figure below.
    In this picture, the qubits live on the edges and there
    are two types of stabilizers: vertex Z stabilizers and
    face X stabilizers.
    Note the weight-2 stabilizers on the boundaries.
    The below lattice is of size 5 by 4.

    .. image:: rotated_planar_2d_code.svg
        :scale: 200 %
        :align: center

    Alternatively, we can consider the the same coordinates,
    qubits and stabilizers but seen on a rotated lattice,
    where the lattice live on vertices,
    and the two types of stabilizers correspond to the color
    of the checkerboard faces as shown below.

    .. image:: rotated_planar_2d_code_2.svg
        :scale: 200 %
        :align: center
    """
    dimension = 2
    deformation_names = ['XZZX', 'XY']

    @property
    def label(self) -> str:
        return 'Rotated Planar {}x{}'.format(*self.size)

    def get_qubit_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly = self.size

        # Qubits along e_x
        for x in range(1, 2*Lx+1, 2):
            for y in range(1, 2*Ly+1, 2):
                coordinates.append((x, y))

        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly = self.size

        # Vertices
        for x in range(2, 2*Lx, 2):
            for y in range(0, 2*Ly+1, 2):
                if (x + y) % 4 == 2:
                    coordinates.append((x, y))

        # Faces
        for x in range(0, 2*Lx+1, 2):
            for y in range(2, 2*Ly, 2):
                if (x + y) % 4 == 0:
                    coordinates.append((x, y))

        return coordinates

    def stabilizer_type(self, location: Tuple):
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location} for a stabilizer")

        x, y = location
        if (x + y) % 4 == 2:
            return 'vertex'
        else:
            return 'face'

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location} for a stabilizer")

        if self.stabilizer_type(location) == 'vertex':
            pauli = 'Z'
        else:
            pauli = 'X'

        delta = [(-1, -1), (-1, 1), (1, -1), (1, 1)]

        operator = dict()
        for d in delta:
            qubit_location = tuple(np.add(location, d))

            if self.is_qubit(qubit_location):
                operator[qubit_location] = pauli

        return operator

    def qubit_axis(self, location):
        x, y = location

        if (x + y) % 4 == 2:
            axis = 'x'
        elif (x + y) % 4 == 0:
            axis = 'y'
        else:
            raise ValueError(f'Location {location} does not correspond'
                             f'to a qubit')

        return axis

    def get_logicals_x(self) -> List[Operator]:
        Lx, Ly = self.size
        logicals = []

        # X operators along first diagonal
        operator: Operator = dict()
        for x in range(1, 2*Lx+1, 2):
            operator[(x, 1)] = 'X'
        logicals.append(operator)

        return logicals

    def get_logicals_z(self) -> List[Operator]:
        Lx, Ly = self.size
        logicals = []

        # Z operators along first diagonal
        operator: Operator = dict()
        for y in range(1, 2*Ly+1, 2):
            operator[(1, y)] = 'Z'
        logicals.append(operator)

        return logicals

    def get_deformation(
        self, location: Tuple,
        deformation_name: str,
        deformation_axis: str = 'y',
        **kwargs
    ) -> Dict:

        if deformation_axis not in ['x', 'y']:
            raise ValueError(f"{deformation_axis} is not a valid "
                             "deformation axis")

        if deformation_name == 'XZZX':
            undeformed_dict = {'X': 'X', 'Y': 'Y', 'Z': 'Z'}
            deformed_dict = {'X': 'Z', 'Y': 'Y', 'Z': 'X'}

            if self.qubit_axis(location) == deformation_axis:
                deformation = deformed_dict
            else:
                deformation = undeformed_dict

        elif deformation_name == 'XY':
            deformation = {'X': 'X', 'Y': 'Z', 'Z': 'Y'}

        else:
            raise ValueError(f"The deformation {deformation_name}"
                             "does not exist")

        return deformation



import pytest
from panqec.codes import Color3DCode
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestColor3DCode(StabilizerCodeTest):

    @pytest.fixture(params=[2, 4])
    def code(self, request):
        return Color3DCode(request.param)



import numpy as np
import pytest
from click.testing import CliRunner

from panqec.cli import cli, read_bias_ratios, read_range_input


@pytest.fixture
def runner():
    """Click CliRunner with isolated file system."""
    _runner = CliRunner()
    with _runner.isolated_filesystem():
        yield _runner
    assert hasattr(_runner, 'invoke')


@pytest.mark.parametrize('arguments', [
    [],
    ['--help'],
    ['--version'],
])
def test_cli_basic(arguments, runner):
    result = runner.invoke(cli, arguments)
    assert result.exit_code == 0


class TestLS:

    def test_ls_all(self, runner):
        result = runner.invoke(cli, ['ls'])
        assert result.exit_code == 0
        assert 'Codes:' in result.output
        assert 'Error Models (Noise):' in result.output
        assert 'Decoders:' in result.output

    def test_ls_codes(self, runner):
        result = runner.invoke(cli, ['ls', 'codes'])
        assert result.exit_code == 0
        assert 'Codes:' in result.output
        assert 'Error Models (Noise):' not in result.output
        assert 'Decoders:' not in result.output

    def test_ls_error_models(self, runner):
        result = runner.invoke(cli, ['ls', 'error_models'])
        assert result.exit_code == 0
        assert 'Codes:' not in result.output
        assert 'Error Models (Noise):' in result.output
        assert 'Decoders:' not in result.output

    def test_ls_decoders(self, runner):
        result = runner.invoke(cli, ['ls', 'decoders'])
        assert result.exit_code == 0
        assert 'Codes:' not in result.output
        assert 'Error Models (Noise):' not in result.output
        assert 'Decoders:' in result.output


def test_read_bias_ratios():
    expected_bias_ratios = [0.5, 1, 3, 10, 30, 100, np.inf]
    eta_string = '0.5,1,3,10,30,100,inf'
    bias_ratios = read_bias_ratios(eta_string)
    assert len(bias_ratios) == len(expected_bias_ratios)
    for eta, expected_eta in zip(bias_ratios, expected_bias_ratios):
        assert eta == expected_eta
        assert type(eta) == type(expected_eta)


@pytest.mark.parametrize('spec,expected_values', [
    ('0:0.6:0.005', np.arange(0, 0.605, 0.005).tolist()),
    ('1,2,3', [1.0, 2.0, 3.0]),
    ('13.21', [13.21]),
    ('1e-2', [0.01]),
])
def test_read_range_input(spec, expected_values):
    values = read_range_input(spec)
    assert len(values) == len(expected_values)
    for value, expected_value in zip(values, expected_values):
        assert value == expected_value
        assert type(value) == type(expected_value)



import functools
from typing import Tuple, Optional
import numpy as np
from panqec.codes import StabilizerCode
from . import BaseErrorModel
from panqec.bpauli import pauli_to_bsf
import random


def fast_choice(options, probs, rng=None):
    """Found on stack overflow to accelerate np.random.choice"""
    if rng is None:
        x = random.random()
    else:
        x = rng.random()
    cum = 0
    for i, p in enumerate(probs):
        cum += p
        if x < cum:
            return options[i]
    return options[-1]


class PauliErrorModel(BaseErrorModel):
    """Pauli channel IID noise model.

    The overal quantum channel takes the form

    .. math::

        \\mathcal{\\rho}(\\rho) =
        (1 - p)\\rho + p(r_X X \\rho X + r_Y Y \\rho Y + r_Z Z \\rho Z)

    where it is required that :math:`r_X + r_Y + r_Z =1`.
    """

    def __init__(
        self,
        r_x: float, r_y: float, r_z: float,
        deformation_name: Optional[str] = None,
        deformation_kwargs: Optional[dict] = None
    ):
        """Initialize Pauli error model at a given rate of X, Y and Z errors,
        i.e. :math:`p_\\sigma = p r_\\sigma` for
        :math:`\\sigma \\in \\{X, Y, Z\\}`,
        :math:`p` the total error rate,
        and :math:`p_\\sigma` the probability of getting
        the error :math:`\\sigma` on each qubit.

        Parameters
        ----------
        r_x : float
            Rate of X errors
        r_y : float
            Rate of Y errors
        r_z : float
            Rate of Z errors
        deformation_name : str, optional
            Name of the Clifford deformation to apply to the noise model.
            The Clifford deformation must be provided in the code class.
        """
        if not np.isclose(r_x + r_y + r_z, 1):
            raise ValueError(
                f'Noise direction ({r_x}, {r_y}, {r_z}) does not sum to 1.0'
            )
        self._direction = r_x, r_y, r_z
        self._deformation_name = deformation_name

        if deformation_kwargs is not None:
            self._deformation_kwargs = deformation_kwargs
        else:
            self._deformation_kwargs = {}

    @property
    def direction(self) -> Tuple[float, float, float]:
        """Rate of X, Y and Z errors, as given when initializing the
        error model

        Returns
        -------
        (r_x, r_y, r_z): Tuple[float]
            Rate of X, Y and Z errors
        """
        return self._direction

    @property
    def label(self):
        label = 'Pauli X{:.4f}Y{:.4f}Z{:.4f}'.format(*self.direction)
        if self._deformation_name:
            label = 'Deformed ' + self._deformation_name + ' ' + label

        return label

    @property
    def params(self) -> dict:
        """List of class arguments (as a dictionary), that can be saved
        and reused to instantiate the same code"""
        return {
            'r_x': self.direction[0],
            'r_y': self.direction[1],
            'r_z': self.direction[2],
            'deformation_name': self._deformation_name,
            'deformation_kwargs': self._deformation_kwargs
        }

    def generate(self, code: StabilizerCode, error_rate: float, rng=None):
        rng = np.random.default_rng() if rng is None else rng

        p_i, p_x, p_y, p_z = self.probability_distribution(code, error_rate)

        error_pauli = ''.join([fast_choice(
            ('I', 'X', 'Y', 'Z'),
            [p_i[i], p_x[i], p_y[i], p_z[i]],
            rng=rng
        ) for i in range(code.n)])

        error = pauli_to_bsf(error_pauli)

        return error

    @functools.lru_cache()
    def probability_distribution(
        self, code: StabilizerCode, error_rate: float
    ) -> Tuple:
        n = code.n
        r_x, r_y, r_z = self.direction

        p: dict = {}
        p['I'] = (1 - error_rate) * np.ones(n)
        p['X'] = (r_x * error_rate) * np.ones(n)
        p['Y'] = (r_y * error_rate) * np.ones(n)
        p['Z'] = (r_z * error_rate) * np.ones(n)

        if self._deformation_name is not None:
            for i in range(code.n):
                deformation = code.get_deformation(
                    code.qubit_coordinates[i], self._deformation_name,
                    **self._deformation_kwargs
                )
                previous_p = {pauli: p[pauli][i] for pauli in ['X', 'Y', 'Z']}
                for pauli in ['X', 'Y', 'Z']:
                    p[pauli][i] = previous_p[deformation[pauli]]

        return p['I'], p['X'], p['Y'], p['Z']



import numpy as np
from panqec.usage import TextPlotter


class TestTextPlotter:

    def test_plot_lissajous_figure(self):
        plotter = TextPlotter(width=79, height=34)
        x = np.linspace(-10, 10, 1001)
        plotter.plot(np.cos(3*x), np.sin(3*x), '.')
        plotter.plot(0.5*np.cos(3*x), 0.5*np.sin(3*x), '@')
        plotter.plot(np.cos(3*x) - np.sin(2*x), np.cos(3*x) + np.sin(2*x), 'o')
        plotter.xlabel('x (m)')
        plotter.ylabel('y (m)')
        plotter.title('Cool title goes here')
        plotter.show()



import pytest
import numpy as np
from panqec.codes import Toric2DCode, Toric3DCode
from panqec.bpauli import bsf_wt
from panqec.decoders import MatchingDecoder
from panqec.error_models import PauliErrorModel
from tests.decoders.decoder_test import DecoderTest


class TestToric2DMatchingDecoder(DecoderTest):

    @pytest.fixture
    def code(self):
        return Toric2DCode(4, 5)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.1
        return MatchingDecoder(code, error_model, error_rate)

    @pytest.mark.parametrize(
        'operator, location',
        [
            ('X', (0, 1)),
            ('Y', (0, 1)),
            ('Z', (0, 1)),
            ('X', (1, 0)),
            ('Y', (1, 0)),
            ('Z', (1, 0)),
        ]
    )
    def test_decode_single_error(self, code, operator, location):
        error_model = PauliErrorModel(1/3, 1/3, 1/3)
        error_rate = 0.5
        decoder = MatchingDecoder(code, error_model, error_rate)
        error = code.to_bsf({
            location: operator
        })
        assert bsf_wt(error) == 1, 'Error should be weight 1'
        syndromes = code.measure_syndrome(error)
        if operator in ['Z', 'X']:
            assert sum(syndromes) == 2, 'Should be 2 syndromes if X or Z error'
        elif operator == 'Y':
            assert sum(syndromes) == 4, 'Should be 4 syndromes if Y error'
        correction = decoder.decode(syndromes)
        assert bsf_wt(correction) == 1, 'Correction should be weight 1'
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0), (
            'Total error should be in code space'
        )

    def test_exception_when_wrong_code(self, code):
        error_model = PauliErrorModel(1/3, 1/3, 1/3)
        error_rate = 0.5
        MatchingDecoder(code, error_model, error_rate)

        code_3d = Toric3DCode(3)
        with pytest.raises(Exception):
            MatchingDecoder(code_3d, error_model, error_rate)



from typing import Tuple, Dict, List
from panqec.codes import StabilizerCode

Operator = Dict[Tuple, str]  # Location to pauli ('X', 'Y' or 'Z')
Coordinates = List[Tuple]  # List of locations


class Color666PlanarCode(StabilizerCode):
    """2D color code on 6,6,6 lattice with open boundaries.

    The overall shape of the lattice is roughly a triangle.

    Parameters
    ----------
    L_x : int
        Number of unit cells in x direction.
    L_y : Optional[int]
        Number of unit cells in y direction.

    Notes
    -----
    The 6,6,6 lattice is a tessallation of red hexagons (6),
    green hexagons (6) and blue hexagons (6).
    Qubits live on the vertices.
    For each colored shape, there are two stabilizer generators,
    one of all X over the qubits on its corners,
    and one of all Z over the qubits on its corners.

    See `EC zoo article <https://errorcorrectionzoo.org/c/color>`_
    for further reading.
    """
    dimension = 2

    @property
    def label(self) -> str:
        return 'Color 6.6.6 Planar {}x{}'.format(*self.size)

    def get_qubit_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly = self.size

        coordinates = []

        stab_coordinates = self.get_stabilizer_coordinates()

        for location in stab_coordinates:
            qubit_coords = list(self.get_stabilizer(location).keys())
            for coord in qubit_coords:
                if coord not in coordinates:
                    coordinates.append(coord)

        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly = self.size

        for x in range(2, 12*Lx+4, 1):
            for y in range(0, min(2*x + 1, 12*Lx - 2*x + 3), 1):
                if (x % 6 == 2 and y % 4 == 0) or (x % 6 == 5 and y % 4 == 2):
                    coordinates.append((x, y, 0))
                    coordinates.append((x, y, 1))

        return coordinates

    def stabilizer_type(self, location: Tuple) -> str:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        x, y, pauli = location
        stab_type = 'face'

        if (x % 6 == 2 and y % 12 == 0) or (x % 6 == 5 and y % 12 == 6):
            stab_type += '-green'
        elif (x % 6 == 5 and y % 12 == 2) or (x % 6 == 2 and y % 12 == 8):
            stab_type += '-blue'
        else:
            stab_type += '-red'

        if pauli == 0:
            stab_type += '-x'
        else:
            stab_type += '-z'

        return stab_type

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location}"
                             "for a stabilizer")

        Lx, Ly = self.size

        if location[-1] == 0:
            pauli = 'X'
        else:
            pauli = 'Z'

        delta: List[Tuple] = [(-1, -2), (1, -2), (2, 0),
                              (1, 2), (-1, 2), (-2, 0)]

        operator: Operator = dict()
        for d in delta:
            qubit_location = (location[0] + d[0], location[1] + d[1])
            x, y = qubit_location

            if x >= 0 and 0 <= y <= min(2*x + 1, 12*Lx - 2*x + 3):
                operator[qubit_location] = pauli

        return operator

    def qubit_axis(self, location: Tuple) -> str:
        x, y = location

        return 'x'

    def get_logicals_x(self) -> List[Operator]:
        Lx, Ly = self.size
        logicals: List[Operator] = []

        # X operators in x direction.
        operator: Operator = dict()
        for x in range(0, 12*Lx+4, 2):
            if self.is_qubit((x, 0)):
                operator[(x, 0)] = 'X'
        logicals.append(operator)

        return logicals

    def get_logicals_z(self) -> List[Operator]:
        Lx, Ly = self.size
        logicals: List[Operator] = []

        # Z operators x direction.
        operator: Operator = dict()
        for x in range(0, 12*Lx+4, 2):
            if self.is_qubit((x, 0)):
                operator[(x, 0)] = 'Z'
        logicals.append(operator)

        return logicals

    def stabilizer_representation(
        self, location: Tuple, rotated_picture=False, json_file=None
    ) -> Dict:
        rep = super().stabilizer_representation(location, rotated_picture)

        Lx, Ly = self.size
        x, y, _ = location

        # We remove the last part of the location (indexing X or Z)
        rep['location'] = (x, y)

        if y == 2*x:
            rep["params"]["vertices"] = [[-1, -2], [1, -2], [2, 0], [1, 2]]

        if y == 12*Lx - 2*x:
            rep["params"]["vertices"] = [[-1, -2], [1, -2], [-1, 2], [-2, 0]]

        if y == 0:
            rep["params"]["vertices"] = [[2, 0], [1, 2], [-1, 2], [-2, 0]]

        return rep



import pytest
from panqec.codes import HollowRhombicCode
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestHollowRhombicCode(StabilizerCodeTest):

    @pytest.fixture(params=[4, 6])
    def code(self, request):
        return HollowRhombicCode(request.param)



import pytest
from itertools import combinations
import numpy as np
from panqec.bpauli import bs_prod, bsf_wt
from panqec.codes import (
    RotatedPlanar3DCode
)
from panqec.decoders import RotatedSweepMatchDecoder, RotatedSweepDecoder3D
from panqec.error_models import PauliErrorModel
from tests.decoders.decoder_test import DecoderTest


class TestRotatedSweepMatchDecoder(DecoderTest):

    @pytest.fixture
    def code(self):
        return RotatedPlanar3DCode(4, 4, 4)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.5
        return RotatedSweepMatchDecoder(code, error_model, error_rate,
                                        max_rounds=4)

    @pytest.mark.parametrize('sweep_direction, diffs', [
        [(+1, 0, +1), [(+1, +1, +1), (+1, -1, +1), (+2, 0, 0)]],
        [(+1, 0, -1), [(+1, +1, -1), (+1, -1, -1), (+2, 0, 0)]],
        [(0, +1, +1), [(+1, +1, +1), (-1, +1, +1), (0, +2, 0)]],
        [(0, +1, -1), [(+1, +1, -1), (-1, +1, -1), (0, +2, 0)]],
        [(-1, 0, +1), [(-1, -1, +1), (-1, +1, +1), (-2, 0, 0)]],
        [(-1, 0, -1), [(-1, -1, -1), (-1, +1, -1), (-2, 0, 0)]],
        [(0, -1, +1), [(-1, -1, +1), (+1, -1, +1), (0, -2, 0)]],
        [(0, -1, -1), [(-1, -1, -1), (+1, -1, -1), (0, -2, 0)]],
    ])
    def test_get_sweep_faces(self, sweep_direction, diffs, decoder):
        vertex = (0, 0, 0)
        expected_x_face, expected_y_face, expected_z_face = [
            tuple(np.array(vertex) + np.array(diff))
            for diff in diffs
        ]
        x_face, y_face, z_face = decoder.sweeper.get_sweep_faces(
            vertex, sweep_direction
        )
        assert x_face == expected_x_face
        assert y_face == expected_y_face
        assert z_face == expected_z_face

    @pytest.mark.parametrize('sweep_direction, diffs', [
        [(+1, 0, +1), [(+1, -1, 0), (+1, +1, 0), (0, 0, +1)]],
        [(+1, 0, -1), [(+1, -1, 0), (+1, +1, 0), (0, 0, -1)]],
        [(0, +1, +1), [(-1, +1, 0), (+1, +1, 0), (0, 0, +1)]],
        [(0, +1, -1), [(-1, +1, 0), (+1, +1, 0), (0, 0, -1)]],
        [(-1, 0, +1), [(-1, +1, 0), (-1, -1, 0), (0, 0, +1)]],
        [(-1, 0, -1), [(-1, +1, 0), (-1, -1, 0), (0, 0, -1)]],
        [(0, -1, +1), [(+1, -1, 0), (-1, -1, 0), (0, 0, +1)]],
        [(0, -1, -1), [(+1, -1, 0), (-1, -1, 0), (0, 0, -1)]],
    ])
    def test_get_sweep_edges(self, sweep_direction, diffs, decoder):
        vertex = (0, 0, 0)
        expected_x_edge, expected_y_edge, expected_z_edge = [
            tuple(np.array(vertex) + np.array(diff))
            for diff in diffs
        ]
        x_edge, y_edge, z_edge = decoder.sweeper.get_sweep_edges(
            vertex, sweep_direction
        )
        assert x_edge == expected_x_edge
        assert y_edge == expected_y_edge
        assert z_edge == expected_z_edge

    @pytest.mark.parametrize(
        'paulis_locations',
        [
            [('X', (3, 3, 1)), ('X', (5, 1, 5)), ('X', (6, 4, 6))],
            [('Z', (3, 3, 1)), ('Z', (5, 1, 5)), ('Z', (6, 4, 6))],
            [('Y', (1, 5, 1)), ('Y', (2, 4, 4)), ('Y', (6, 4, 4))],
            [('X', (1, 1, 1)), ('X', (1, 3, 1))],
            [('X', (1, 1, 1)), ('X', (3, 1, 1))],
            [('X', (2, 0, 2)), ('X', (2, 0, 4))],
            [('Z', (1, 1, 1)), ('Z', (1, 3, 1))],
            [('Z', (1, 1, 1)), ('Z', (3, 1, 1))],
            [('Z', (2, 0, 2)), ('Z', (2, 0, 4))],
            [('Y', (1, 1, 1)), ('Y', (1, 3, 1))],
            [('Y', (1, 1, 1)), ('Y', (3, 1, 1))],
            [('Y', (2, 0, 2)), ('Y', (2, 0, 4))],
            [('Z', (1, 1, 1))],
        ],
        ids=[
            'X_bulk', 'Z_bulk', 'Y_bulk',
            'X_boundary_x', 'X_boundary_y', 'X_boundary_z',
            'Z_boundary_x', 'Z_boundary_y', 'Z_boundary_z',
            'Y_boundary_x', 'Y_boundary_y', 'Y_boundary_z',
            'Z_corner'
        ]
    )
    def test_decode_many_errors(self, decoder, code, paulis_locations):
        error = code.to_bsf({
            location: pauli
            for pauli, location in paulis_locations
        })
        assert bsf_wt(error) == len(paulis_locations)

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_undecodable_error(self, decoder, code):
        locations = [
            (x, y, z) for x, y, z in code.qubit_coordinates
        ]
        assert len(locations) > 0
        error = code.to_bsf({
            location: 'Z'
            for location in locations
        })
        assert bsf_wt(error) == len(locations)

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.any(total_error)

    def test_decode_many_codes_and_errors_with_same_decoder(self):

        codes_sites = [
            (RotatedPlanar3DCode(3, 3, 3), (3, 3, 3)),
            (RotatedPlanar3DCode(4, 4, 4), (5, 5, 5)),
            (RotatedPlanar3DCode(5, 5, 5), (3, 3, 3)),
        ]

        error_model = PauliErrorModel(1/3, 1/3, 1/3)
        error_rate = 0.5

        for code, site in codes_sites:
            decoder = RotatedSweepMatchDecoder(code, error_model, error_rate,
                                               max_rounds=4)
            error = code.to_bsf({site: 'Z'})
            syndrome = code.measure_syndrome(error)
            correction = decoder.decode(syndrome)
            total_error = (error + correction) % 2
            assert np.all(code.measure_syndrome(total_error) == 0)


class TestSweepMatch3x3x3(DecoderTest):
    """Test cases found to be failing on the GUI."""

    @pytest.fixture
    def code(self):
        return RotatedPlanar3DCode(3, 3, 3)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.5
        return RotatedSweepMatchDecoder(code, error_model, error_rate)

    @pytest.mark.parametrize('locations', [
        [('Z', (1, 1, 1)), ('Z', (3, 3, 1)), ('Z', (5, 5, 1))],
        [('Z', (1, 1, 3)), ('Z', (3, 3, 3)), ('Z', (5, 5, 3))],
        [('Z', (3, 5, 3)), ('Z', (5, 3, 3)), ('Z', (5, 1, 3))],
    ], ids=[
        'up_left_horizontal_bottom',
        'up_left_horizontal_top',
        'down_right_horizontal'
    ])
    def test_errors_spanning_boundaries_fail(self, code, decoder, locations):
        error = code.to_bsf({
            location: pauli
            for pauli, location in locations
        })
        assert bsf_wt(error) == len(locations)

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert not np.all(
            code.measure_syndrome(total_error) == 0
        ), 'Total error in codespace when it should not be'


class TestSweepMatch4x4x3(DecoderTest):
    """Test cases found to be failing on the GUI."""

    @pytest.fixture
    def code(self):
        return RotatedPlanar3DCode(5, 5, 5)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.5
        return RotatedSweepMatchDecoder(code, error_model, error_rate,
                                        max_rounds=4)

    @pytest.mark.parametrize('locations', [
        [
            ('Z', (1, 1, 5)), ('Z', (3, 3, 5)), ('Z', (5, 5, 5)),
            ('Z', (7, 5, 5)), ('Z', (9, 3, 5)),
        ],
        [
            ('Z', (1, 1, 3)), ('Z', (1, 3, 5)),
        ],
        [
            ('Z', (1, 1, 5)), ('Z', (1, 3, 5)),
        ],
        [
            ('Z', (1, 3, 1)), ('Z', (3, 1, 5)),
        ],
        [
            ('Z', (1, 3, 1)), ('Z', (4, 2, 4)),
        ],
        [
            ('Z', (1, 3, 3)), ('Z', (3, 1, 5)),
        ],
    ], ids=[
        'arthurs_example',
        'weight_2_Z_error_1',
        'weight_2_Z_error_2',
        'weight_2_Z_error_3',
        'weight_2_Z_error_4',
        'weight_2_Z_error_5',
    ])
    def test_gui_examples(self, code, decoder, locations):
        error = code.to_bsf({
            location: pauli
            for pauli, location in locations
        })
        assert bsf_wt(error) == len(locations)

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0), (
            'Total error not in codespace'
        )

        assert np.all(bs_prod(code.logicals_x, total_error) == 0), (
            'Total error anticommutes with logical X'
        )
        assert np.all(bs_prod(code.logicals_z, total_error) == 0), (
            'Total error anticommutes with logical Z'
        )

    @pytest.mark.parametrize('locations', [
        [
            ('Z', (1, 1, 1)), ('Z', (3, 3, 1)), ('Z', (5, 5, 1)),
            ('Z', (7, 7, 1)), ('Z', (9, 9, 1))
        ],
        [
            ('Z', (1, 1, 5)), ('Z', (3, 3, 5)), ('Z', (5, 5, 5)),
            ('Z', (7, 7, 5)), ('Z', (9, 9, 5))
        ],
        [
            ('Z', (3, 9, 5)), ('Z', (5, 7, 5)), ('Z', (7, 5, 5)),
            ('Z', (9, 3, 5)), ('Z', (9, 1, 5))
        ],
    ], ids=[
        'up_left_horizontal_bottom',
        'up_left_horizontal_top',
        'down_right_horizontal',
    ])
    def test_errors_spanning_boundaries(self, code, decoder, locations):
        error = code.to_bsf({
            location: pauli
            for pauli, location in locations
        })
        assert bsf_wt(error) == len(locations)

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert not np.all(
            code.measure_syndrome(total_error) == 0
        ), 'Total error in codespace when it should not be'


class TestSweepCorners(DecoderTest):
    """Test 1-qubit errors on corners fully correctable."""

    @pytest.fixture
    def code(self):
        return RotatedPlanar3DCode(5, 5, 3)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.5
        return RotatedSweepMatchDecoder(code, error_model, error_rate)

    @pytest.mark.parametrize('location', [
        (1, 3, 5),
        (3, 5, 5),
        (5, 7, 5),
        (7, 9, 5),
        (9, 9, 5)
    ])
    def test_sweep_errors_on_extreme_layer(self, code, decoder, location):
        error = code.to_bsf({
            location: 'Z'
        })
        assert bsf_wt(error) == 1

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0), (
            'Total error not in codespace'
        )

        assert np.all(bs_prod(code.logicals_x, total_error) == 0), (
            'Total error anticommutes with logical X'
        )
        assert np.all(bs_prod(code.logicals_z, total_error) == 0), (
            'Total error anticommutes with logical Z'
        )

    @pytest.mark.parametrize('pauli', ['X', 'Y', 'Z'])
    def test_all_1_qubit_errors_correctable(self, code, decoder, pauli):
        uncorrectable_locations = []
        for location in code.qubit_coordinates:
            error = code.to_bsf({
                location: pauli
            })
            assert bsf_wt(error) == 1

            syndrome = code.measure_syndrome(error)
            assert np.any(syndrome != 0)

            correction = decoder.decode(syndrome)
            total_error = (error + correction) % 2
            assert np.all(code.measure_syndrome(total_error) == 0)

            correctable = True
            if np.any(bs_prod(code.logicals_x, total_error) != 0):
                correctable = False
            if np.all(bs_prod(code.logicals_z, total_error) != 0):
                correctable = False
            if not correctable:
                uncorrectable_locations.append(location)

        assert len(uncorrectable_locations) == 0, (
            f'Found {len(uncorrectable_locations)} uncorrectable weight-1 '
            f'{pauli} errors'
        )

    @pytest.mark.slow
    def test_all_2_qubit_errors_correctable(self, code, decoder):
        pauli = 'Z'
        weight = 2
        error_locations = combinations(list(code.qubit_index), weight)
        uncorrectable_error_locations = []
        for locations in error_locations:
            error = code.to_bsf({
                location: pauli
                for location in locations
            })
            assert bsf_wt(error) == len(locations)

            syndrome = code.measure_syndrome(error)
            assert np.any(syndrome != 0)

            correction = decoder.decode(syndrome)
            total_error = (error + correction) % 2
            assert np.all(code.measure_syndrome(total_error) == 0)

            correctable = True
            if np.any(bs_prod(code.logicals_x, total_error) != 0):
                correctable = False
            if np.all(bs_prod(code.logicals_z, total_error) != 0):
                correctable = False
            if not correctable:
                uncorrectable_error_locations.append(locations)

        assert len(uncorrectable_error_locations) == 0, (
            f'Found {len(uncorrectable_error_locations)} uncorrectable '
            f'weight-2 Z errors'
        )


class TestRotatedSweepDecoder3D(DecoderTest):

    @pytest.fixture
    def code(self):
        return RotatedPlanar3DCode(3, 3, 3)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.5
        return RotatedSweepDecoder3D(code, error_model, error_rate)

    @pytest.fixture
    def allowed_paulis(self):
        return ['Z']

    @pytest.mark.parametrize(
        'vertex,sweep_direction,sweep_faces,sweep_edges',
        [
            (
                (4, 6, 5),
                (1, 0, -1),
                [(5, 7, 4), (5, 5, 4), (6, 6, 5)],
                [(5, 5, 5), (5, 7, 5), (4, 6, 4)],
            )
        ]
    )
    def test_sweep_faces_edges_top_boundary(
        self, code, decoder, vertex, sweep_direction, sweep_faces, sweep_edges
    ):
        vertex = (4, 6, 5)
        sweep_direction = (1, 0, -1)
        assert code.stabilizer_type(vertex) == 'vertex'
        x_face, y_face, z_face = decoder.get_sweep_faces(
            vertex, sweep_direction
        )
        x_edge, y_edge, z_edge = decoder.get_sweep_edges(
            vertex, sweep_direction
        )
        assert [x_face, y_face, z_face] == sweep_faces
        assert [x_edge, y_edge, z_edge] == sweep_edges

    def test_sweep_touches_all_faces_and_qubits(self, code, decoder):
        touched_edges = []
        touched_faces = []
        sweep_directions = [
            (1, 0, 1), (1, 0, -1),
            (0, 1, 1), (0, 1, -1),
            (-1, 0, 1), (-1, 0, -1),
            (0, -1, 1), (0, -1, -1),
        ]
        vertices = [
            location
            for location in code.stabilizer_coordinates
            if code.stabilizer_type(location) == 'vertex'
        ]
        face_index = {
            location: index
            for index, location in enumerate(code.stabilizer_coordinates)
            if code.stabilizer_type(location) == 'face'
        }
        for sweep_direction in sweep_directions:
            for vertex in vertices:
                x_face, y_face, z_face = decoder.get_sweep_faces(
                    vertex, sweep_direction
                )
                x_edge, y_edge, z_edge = decoder.get_sweep_edges(
                    vertex, sweep_direction
                )

                faces_valid = tuple(
                    face in face_index
                    for face in [x_face, y_face, z_face]
                )
                edges_valid = tuple(
                    edge in code.qubit_coordinates
                    for edge in [x_edge, y_edge, z_edge]
                )
                if all(faces_valid) and all(edges_valid):
                    x_face_bsf = code.stabilizer_matrix[face_index[x_face]]
                    y_face_bsf = code.stabilizer_matrix[face_index[y_face]]
                    z_face_bsf = code.stabilizer_matrix[face_index[z_face]]

                    error = dict()
                    error[x_edge] = 'Z'
                    x_edge_bsf = code.to_bsf(error)
                    assert np.all(bs_prod(x_edge_bsf, x_face_bsf) == 0)
                    assert np.any(bs_prod(x_edge_bsf, y_face_bsf) == 1)
                    assert np.any(bs_prod(x_edge_bsf, z_face_bsf) == 1)

                    error = dict()
                    error[y_edge] = 'Z'
                    y_edge_bsf = code.to_bsf(error)
                    assert np.all(bs_prod(y_edge_bsf, y_face_bsf) == 0)
                    assert np.any(bs_prod(y_edge_bsf, x_face_bsf) == 1)
                    assert np.any(bs_prod(y_edge_bsf, z_face_bsf) == 1)

                    error = dict()
                    error[z_edge] = 'Z'
                    z_edge_bsf = code.to_bsf(error)
                    assert np.any(bs_prod(z_edge_bsf, z_face_bsf) == 0)
                    assert np.any(bs_prod(z_edge_bsf, x_face_bsf) == 1)
                    assert np.any(bs_prod(z_edge_bsf, y_face_bsf) == 1)

                    touched_edges.append(x_edge)
                    touched_edges.append(y_edge)
                    touched_edges.append(z_edge)

                    touched_faces.append(x_face)
                    touched_faces.append(y_face)
                    touched_faces.append(z_face)

        assert set(code.qubit_coordinates) == set(touched_edges)
        assert set(face_index.keys()) == set(touched_faces)

    def test_flip_edge(self, code, decoder):
        for edge in code.qubit_coordinates:
            signs = {
                location: 0
                for location in code.stabilizer_coordinates
                if code.stabilizer_type(location) == 'face'
            }
            signs = decoder.get_initial_state(
                np.zeros(code.stabilizer_matrix.shape[0], dtype=np.uint)
            )
            decoder.flip_edge(edge, signs)
            sign_flip_syndrome = signs

            error = code.to_bsf({
                edge: 'Z'
            })
            pauli_syndrome = code.measure_syndrome(error)

            assert np.all(pauli_syndrome == sign_flip_syndrome)



import pytest
from panqec.codes import RotatedPlanar2DCode
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestRotatedPlanar2DCode(StabilizerCodeTest):

    @pytest.fixture(params=[(2, 2), (3, 3), (2, 3)])
    def code(self, request):
        return RotatedPlanar2DCode(*request.param)



"""
Module to deal with binary sparse matrices, in the compressed sparse row (CSR)
format.

It is mostly a wrapper around scipy.sparse, making sure that everything is in
the correct format and is done as efficiently as possible.
"""

from scipy import sparse
from scipy.sparse import csr_matrix
import numpy as np


def zero_row(n_cols: int):
    """Create a zero sparse row in the csr format"""

    return csr_matrix((np.array([]), (np.array([]), np.array([]))),
                      shape=(1, n_cols), dtype='uint8')


def zero_matrix(shape):
    """Create a zero sparse matrix in the csr format"""

    return csr_matrix((np.array([]), (np.array([]), np.array([]))),
                      shape=shape, dtype='uint8')


def empty_row(n_cols: int):
    """Create an empty sparse row in the csr format"""

    return csr_matrix(([], ([], [])), shape=(0, n_cols), dtype='uint8')


def from_array(array):
    if isinstance(array, list):
        array = np.array(array)
    sparse_matrix = csr_matrix(array, dtype='uint8')
    return sparse_matrix


def to_array(matrix):
    if isinstance(matrix, np.ndarray):
        return matrix
    else:
        return matrix.toarray()


def is_empty(matrix):
    return matrix.shape[0] == 0


def is_sparse(matrix):
    return type(matrix) == csr_matrix


def is_one(index: int, row_matrix):
    """Return True if row_matrix[index] is nonzero"""

    return index in row_matrix.indices


def insert_mod2(index: int, row_matrix):
    """Insert 1 at 'index' in the row matrix, modulo 2"""

    # Check that matrix is a row matrix
    if len(row_matrix.shape) != 2 or row_matrix.shape[0] != 1:
        raise ValueError("The input should be a row matrix,"
                         f"not a {row_matrix.shape}-matrix")

    # If matrix[index] is not zero, put it to zero
    if index in row_matrix.indices:
        row_matrix.indices = np.setdiff1d(row_matrix.indices, [index])
    else:
        row_matrix.indices = np.append(row_matrix.indices, [index])

    row_matrix.data = np.ones(len(row_matrix.indices), dtype='uint8')
    row_matrix.indptr = np.array([0, len(row_matrix.indices)])


def vstack(matrices):
    return sparse.vstack(matrices, format='csr', dtype='uint8')


def hstack(matrices):
    return sparse.hstack(matrices, format='csr', dtype='uint8')


def hsplit(matrix):
    """Split a row matrix into two row matrices with half the number
    of columns"""

    if matrix.shape[1] % 2 != 0:
        raise ValueError("Matrix should have an even number of columns,"
                         f"not {matrix.shape[1]}")

    # Fast version for row matrices
    if matrix.shape[0] == 1:
        indices = np.array(matrix.indices)
        n = matrix.shape[1] // 2

        a_indices = indices[indices < n]
        b_indices = indices[indices >= n] - n

        a_indptr = np.array([0, len(a_indices)])
        b_indptr = np.array([0, len(b_indices)])

        a_data = np.ones(len(a_indices), dtype='uint8')
        b_data = np.ones(len(b_indices), dtype='uint8')

        a = csr_matrix((a_data, a_indices, a_indptr),
                       shape=(1, n), dtype='uint8')
        b = csr_matrix((b_data, b_indices, b_indptr),
                       shape=(1, n), dtype='uint8')
    else:
        n = matrix.shape[1] // 2
        a = matrix[:, :n]
        b = matrix[:, n:]

    return a, b


def dot(a, b):
    """Dot product between two row vectors a and b"""

    # Turn both inputs to sparse format (in case one is not)
    if not is_sparse(a):
        a = from_array(a)
    if not is_sparse(b):
        b = from_array(b)

    # Check if the dimensions are correct
    if a.shape[0] != 1 or b.shape[0] != 1:
        raise ValueError("Dot product only implemented for row vectors,"
                         f"not {a.shape} and {b.shape}")
    if a.shape[1] != b.shape[1]:
        raise ValueError(f"Dimensions {a.shape} and {b.shape} don't agree"
                         "for dot product")

    n_common_ones = len(np.intersect1d(a.indices, b.indices))
    dot_product = int(n_common_ones % 2)

    return dot_product


def equal(a, b):
    """Test if two matrices are equal, or if a matrix equal a unique
    number everywhere"""

    # Matrix equality
    if isinstance(a, csr_matrix) and isinstance(b, csr_matrix):
        return np.all(a.shape == b.shape) and ((a != b).nnz == 0)

    # Matrix and number equality: a should be the int and b the matrix
    if isinstance(b, int) and isinstance(a, csr_matrix):
        a, b = b, a

    if isinstance(a, int) and isinstance(b, csr_matrix):
        if a == 0:
            return (b.nnz == 0)
        else:
            return (len(b.data) == np.product(b.shape)) and np.all(b.data == a)
    else:
        raise TypeError("Equality not supported between"
                        f"{type(a)} and {type(b)}")



"""
API for running simulations.
"""
from abc import ABCMeta, abstractmethod
from json import JSONDecodeError
import datetime
import os
import numpy as np
from panqec.codes import StabilizerCode
from panqec.error_models import BaseErrorModel
from panqec.utils import load_json, save_json


class BaseSimulation(metaclass=ABCMeta):
    """Quantum Error Correction Simulation."""

    start_time: datetime.datetime
    code: StabilizerCode
    error_model: BaseErrorModel
    label: str
    _results: dict = {}
    rng = None

    def __init__(
        self,
        code: StabilizerCode,
        error_model: BaseErrorModel,
        compress: bool = True,
        verbose=True,
        rng=None
    ):
        self.code = code
        self.error_model = error_model
        self.compress = compress
        self.verbose = verbose
        self.rng = rng
        self.label = 'results'

        self._results = {
            'n_runs': 0,
            'wall_time': 0,
        }
        self._inputs = {
            'code': {
                'name': self.code.id,
                'parameters': self.code.params,
                'n': self.code.n,
                'k': self.code.k,
                'd': self.code.d,
            },
            'error_model': {
                'name': self.error_model.id,
                'parameters':  self.error_model.params
            }
        }

    @property
    def wall_time(self):
        return self._results['wall_time']

    @property
    def n_results(self):
        return self._results['n_runs']

    @property
    def results(self):
        res = self._results
        return res

    @property
    def file_name(self) -> str:
        if self.compress:
            extension = '.json.gz'
        else:
            extension = '.json'
        file_name = self.label + extension
        return file_name

    def run(self, n_runs: int):
        self.start_time = datetime.datetime.now()

        self._run(n_runs)

        finish_time = datetime.datetime.now() - self.start_time
        self._results['wall_time'] += finish_time.total_seconds()

    def _find_current_simulation(self, data: list) -> dict:
        for sim in data:
            if sim['inputs'] == self._inputs:
                return sim
        return {}

    def load_results(self, output_file: str):
        """Load previously written results from directory."""

        try:
            if os.path.isfile(output_file):
                data = load_json(output_file)
                data_simulation = self._find_current_simulation(data)

                if data_simulation != {}:
                    self.load_results_from_dict(data_simulation)

        except JSONDecodeError as err:
            print(f'Error loading existing results file {output_file}')
            print('Starting this from scratch')
            print(err)

    def load_results_from_dict(self, data):
        for key in self._results.keys():
            if key in data['results'].keys():
                self._results[key] = data['results'][key]
                if (
                    isinstance(self.results[key], list)
                    and len(self.results[key]) > 0
                    and isinstance(self._results[key][0], list)
                ):
                    self._results[key] = [
                        np.array(array_value)
                        for array_value in self._results[key]
                    ]

    def get_results_to_save(self):
        data = {
            'results': self._results,
            'inputs': self._inputs
        }

        return data

    def save_results(self, output_file: str):
        """Save results to directory."""
        data = self.get_results_to_save()
        save_json(output_file, data)

    @abstractmethod
    def get_results(self):
        pass

    @abstractmethod
    def _run(self, n_runs: int):
        pass

    def postprocess(self):
        pass



import os
import json
import pytest
import numpy as np
from panqec.codes import RotatedToric3DCode
from panqec.bpauli import apply_deformation
from panqec.error_models import PauliErrorModel
from panqec.decoders import BeliefPropagationOSDDecoder
from panqec.bsparse import to_array

from tests.codes.stabilizer_code_test import (
    StabilizerCodeTestWithCoordinates, StabilizerCodeTest
)


class RotatedToric3DCodeTest(StabilizerCodeTestWithCoordinates):
    code_class = RotatedToric3DCode


class TestRotatedToric3DCode2x2x2(RotatedToric3DCodeTest):
    size = (2, 2, 2)
    expected_plane_edges_xy = [
        (1, 1), (3, 1),
        (1, 3), (3, 3),
    ]
    expected_vertical_faces_xy = [
        (1, 1), (1, 3),
        (3, 1), (3, 3),
    ]
    expected_plane_faces_xy = [
        (2, 2), (4, 4)
    ]
    expected_plane_vertices_xy = [
        (4, 2), (2, 4)
    ]
    expected_plane_z = [1, 3]
    expected_vertical_z = [2]


class TestRotatedToric3DCode3x2x2(RotatedToric3DCodeTest):
    size = (3, 2, 2)
    expected_plane_edges_xy = [
        (1, 1), (3, 1), (5, 1),
        (1, 3), (3, 3), (5, 3),
    ]
    expected_vertical_faces_xy = [
        (3, 1), (3, 3),
        (5, 1), (5, 3)
    ]
    expected_plane_faces_xy = [
        (2, 2), (4, 4), (6, 2),
    ]
    expected_plane_vertices_xy = [
        (4, 2), (2, 4), (6, 4)
    ]
    expected_plane_z = [1, 3]
    expected_vertical_z = [2]


@pytest.mark.skip(reason='odd by odd')
class TestRotatedToric3DCode3x3x3(RotatedToric3DCodeTest):
    size = (3, 3, 3)
    expected_plane_edges_xy = [
        (1, 1), (3, 1), (5, 1),
        (1, 3), (3, 3), (5, 3),
        (1, 5), (3, 5), (5, 5),
    ]
    expected_vertical_faces_xy = [
        (3, 3), (3, 5),
        (5, 3), (5, 5),
    ]
    expected_plane_faces_xy = [
        (2, 2), (6, 2),
        (4, 4),
        (2, 6), (6, 6),
    ]
    expected_plane_vertices_xy = [
        (4, 2),
        (2, 4), (6, 4),
        (4, 6),
    ]
    expected_plane_z = [1, 3, 5, 7]
    expected_vertical_z = [2, 4, 6]


class TestRotatedToric3DCode4x3x3(RotatedToric3DCodeTest):
    size = (4, 3, 3)
    expected_plane_edges_xy = [
        (1, 1), (3, 1), (5, 1), (7, 1),
        (1, 3), (3, 3), (5, 3), (7, 3),
        (1, 5), (3, 5), (5, 5), (7, 5),
    ]
    expected_vertical_faces_xy = [
        (1, 3), (1, 5),
        (3, 3), (3, 5),
        (5, 3), (5, 5),
        (7, 3), (7, 5),
    ]
    expected_plane_faces_xy = [
        (2, 2), (6, 2),
        (4, 4), (8, 4),
        (2, 6), (6, 6),
    ]
    expected_plane_vertices_xy = [
        (4, 2), (8, 2),
        (2, 4), (6, 4),
        (4, 6), (8, 6),
    ]
    expected_plane_z = [1, 3, 5]
    expected_vertical_z = [2, 4]


class TestRotatedToric3DCode3x4x3(RotatedToric3DCodeTest):
    size = (3, 4, 3)
    expected_plane_edges_xy = [
        (1, 1), (1, 3), (1, 5), (1, 7),
        (3, 1), (3, 3), (3, 5), (3, 7),
        (5, 1), (5, 3), (5, 5), (5, 7),
    ]
    expected_vertical_faces_xy = [
        (3, 1), (3, 3), (3, 5), (3, 7),
        (5, 1), (5, 3), (5, 5), (5, 7),
    ]
    expected_plane_faces_xy = [
        (2, 2), (2, 6),
        (4, 4), (4, 8),
        (6, 2), (6, 6),
    ]
    expected_plane_vertices_xy = [
        (2, 4), (2, 8),
        (4, 2), (4, 6),
        (6, 4), (6, 8),
    ]
    expected_plane_z = [1, 3, 5]
    expected_vertical_z = [2, 4]


class TestRotatedToric3DCode4x4x4(RotatedToric3DCodeTest):
    size = (4, 4, 4)
    expected_plane_edges_xy = [
        (1, 1), (1, 3), (1, 5), (1, 7),
        (3, 1), (3, 3), (3, 5), (3, 7),
        (5, 1), (5, 3), (5, 5), (5, 7),
        (7, 1), (7, 3), (7, 5), (7, 7),
    ]
    expected_vertical_faces_xy = [
        (1, 1), (1, 3), (1, 5), (1, 7),
        (3, 1), (3, 3), (3, 5), (3, 7),
        (5, 1), (5, 3), (5, 5), (5, 7),
        (7, 1), (7, 3), (7, 5), (7, 7),
    ]
    expected_plane_faces_xy = [
        (2, 2), (2, 6),
        (4, 4), (4, 8),
        (6, 2), (6, 6),
        (8, 4), (8, 8),
    ]
    expected_plane_vertices_xy = [
        (2, 4), (2, 8),
        (4, 2), (4, 6),
        (6, 4), (6, 8),
        (8, 2), (8, 6),
    ]
    expected_plane_z = [1, 3, 5, 7]
    expected_vertical_z = [2, 4, 6]


class TestDeformedRotatedToric3DCode(StabilizerCodeTest):

    @pytest.fixture
    def code(self):
        code = RotatedToric3DCode(5, 6, 5)
        code.deform('XZZX')

        return code


class TestRotatedToric3DDeformation:

    def test_deformation_index(self):
        code = RotatedToric3DCode(3, 4, 4)
        all_deformations = np.array([
            code.get_deformation(loc, 'XZZX', deformation_axis='x')
            for loc in code.qubit_coordinates
        ])
        no_deformation = {'X': 'X', 'Y': 'Y', 'Z': 'Z'}
        deformation_index = np.where(all_deformations != no_deformation)[0]

        deformation_sites = sorted(
            [
                code.qubit_coordinates[i]
                for i in deformation_index
            ],
            key=lambda x: x[::-1]
        )
        assert all(z % 2 == 1 for x, y, z in deformation_sites)
        expected_sites_plane = [
            (1, 1), (5, 1), (3, 3), (1, 5), (5, 5), (3, 7)
        ]
        expected_sites = []
        for z in [1, 3, 5, 7]:
            expected_sites += [(x, y, z) for x, y in expected_sites_plane]
        expected_sites.sort(key=lambda x: x[::-1])
        assert deformation_sites == expected_sites

    @pytest.mark.skip(reason='analytics')
    def test_stabilizer_matrix(self):
        """Run this to generate stabilizer matrix for SageMath.

        The data will be saved as json files in the temp directory.
        The data can then be used as input for a SageMath notebook
        to find the lowest-weight Z-only logical operators by brute force.
        """
        project_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        entries = []
        for size in [3, 4, 5, 6, 7, 8, 9, 10]:
            L_x, L_y, L_z = size, size + 1, size
            print(L_x, L_y, L_z)
            code = RotatedToric3DCode(L_x, L_y, L_z)
            out_json = os.path.join(
                project_dir, 'temp', 'rotated_toric_3d_test.json'
            )
            all_deformations = np.array([
                code.get_deformation(loc, 'XZZX', deformation_axis='x')
                for loc in code.qubit_coordinates
            ])
            no_deformation = {'X': 'X', 'Y': 'Y', 'Z': 'Z'}
            deformation_idx = np.where(all_deformations != no_deformation)[0]

            entries.append({
                'size': [L_x, L_y, L_z],
                'coords': code.qubit_coordinates,
                'undeformed': {
                    'n': int(code.n),
                    'k': int(code.k),
                    'd': int(code.d),
                    'stabilizers': to_array(code.stabilizer_matrix).tolist(),
                    'logicals_x': to_array(code.logicals_x).tolist(),
                    'logicals_z': to_array(code.logicals_z).tolist(),
                },
                'deformed': {
                    'n': int(code.n),
                    'k': int(code.k),
                    'd': int(code.d),
                    'stabilizers': apply_deformation(
                        deformation_idx,
                        to_array(code.stabilizer_matrix)
                    ).tolist(),
                    'logicals_x': apply_deformation(
                        deformation_idx,
                        to_array(code.logicals_x)
                    ).tolist(),
                    'logicals_z': to_array(apply_deformation(
                        deformation_idx,
                        to_array(code.logicals_z)
                    )).tolist(),
                }
            })
        with open(out_json, 'w') as f:
            json.dump({
                'entries': entries
            }, f)


class TestBPOSDOnRotatedToric3DCodeOddTimesEven:

    @pytest.mark.parametrize('pauli', ['X', 'Y', 'Z'])
    def test_decode_single_qubit_error_bposd(self, pauli):
        code = RotatedToric3DCode(3, 4, 3)
        error_model = PauliErrorModel(
            1/3, 1/3, 1/3,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )
        error_rate = 0.1
        decoder = BeliefPropagationOSDDecoder(code, error_model, error_rate)

        failing_cases = []
        for site in code.qubit_coordinates:
            error_pauli = dict()
            error_pauli[site] = pauli
            error = code.to_bsf(error_pauli)
            syndrome = code.measure_syndrome(error)
            correction = decoder.decode(syndrome)
            total_error = (error + correction) % 2
            if not np.all(code.measure_syndrome(total_error) == 0):
                failing_cases.append(site)
        n_failing = len(failing_cases)
        max_show = 100
        if n_failing > max_show:
            failing_cases_show = ', '.join(map(str, failing_cases[:max_show]))
            end_part = f'...{n_failing - max_show} more'
        else:
            failing_cases_show = ', '.join(map(str, failing_cases))
            end_part = ''

        assert n_failing == 0, (
            f'Failed decoding {n_failing} {pauli} errors at '
            f'{failing_cases_show} {end_part}'
        )



"""
Routines for checking usage and progress.

Works by reading usage txt files produced by the `monitor` command.
"""

from typing import List, Dict, Tuple, Any, Optional, Union
import os
import re
from glob import glob
import shutil
import numpy as np
import pandas as pd


def summarize_usage(logs_dirs: List[str]):
    print(f'Checking usage in {len(logs_dirs)} directories')
    usage_df_list = []
    for logs_dir in logs_dirs:
        jobs = get_jobs(logs_dir)
        usage = get_usage(logs_dir, jobs)
        usage_df_list.append(usage)
    if usage_df_list:
        usage_df = pd.concat(usage_df_list, axis=0)
        time_df = get_time_df(usage_df)
        plot_usage(time_df)
    else:
        print('No usage logs found')


def plot_usage(time_df):
    plotter = TextPlotter(height=20)
    plotter.fill(time_df['hour'], time_df['cores_avail'], '.', label='Avail')
    plotter.fill(time_df['hour'], time_df['cores_used'], '#', label='Used')
    plotter.ylabel('CPU cores')
    plotter.title('Resource usage over time')
    plotter.legend()
    plotter.show()

    plotter = TextPlotter(height=20)
    plotter.fill(time_df['hour'], time_df['ram_avail'], '.', label='Avail')
    plotter.fill(time_df['hour'], time_df['ram_used'], '#', label='Used')
    plotter.xlabel('Time (h)')
    plotter.ylabel('RAM (GiB)')
    plotter.legend()
    plotter.show()


def get_jobs(logs_dir):
    jobs_list = []
    usage_files = glob(os.path.join(logs_dir, 'usage_*.txt'))

    for usage_file in usage_files:
        file_name = os.path.split(usage_file)[-1]
        match = re.search(r'usage_(\d+)_(\d+).txt', file_name)
        node_id = match.group(1)
        array_index = match.group(2)
        filter_file = os.path.join(
            logs_dir, f'filter_{node_id}_{array_index}.txt'
        )
        if not os.path.isfile(filter_file):
            filter_file = None
        jobs_list.append({
            'node_id': node_id,
            'array_index': array_index,
            'usage_file': usage_file,
            'filter_file': filter_file,
        })
    jobs = pd.DataFrame(jobs_list)

    jobs['array_index'] = jobs['array_index'].astype('int')
    jobs = jobs.sort_values(by='array_index').reset_index(drop=True)
    return jobs


def get_usage(logs_dir, jobs):
    usage_pattern = re.compile(
        r'^(?P<date>[\d\-\:\.\s]+) '
        r'CPU usage (?P<cpu_pc>[\d\.]+)% \((?P<n_cores>\d+) cores\) '
        r'RAM (?P<ram_pc>[\d.]+)% \((?P<tot_mem>[\d\.]+) GiB tot\)'
    )
    usage_list = []
    for _, job in jobs.iterrows():
        with open(job['usage_file']) as f:
            lines = f.readlines()
        for line in lines:
            usage_match = re.search(usage_pattern, line)
            if usage_match:
                datapoint = {
                    k: usage_match.group(k)
                    for k in usage_pattern.groupindex
                }
                datapoint['node_id'] = job['node_id']
                usage_list.append(datapoint)
    usage = pd.DataFrame(usage_list)
    usage['date'] = pd.to_datetime(usage['date'])
    for column, dtype in [
        ('cpu_pc', float), ('n_cores', int),
        ('ram_pc', float), ('tot_mem', float)
    ]:
        usage[column] = usage[column].astype(dtype)
    usage['cores'] = usage['cpu_pc']*usage['n_cores']/100
    usage['ram'] = usage['ram_pc']*usage['tot_mem']/100
    return usage


def get_time_df(usage):
    date_ranges = pd.date_range(
        usage.date.min(), usage.date.max(), periods=100
    )
    cores_used = [0]
    cores_avail = [0]
    ram_used = [0]
    ram_avail = [0]
    for i in range(len(date_ranges) - 1):
        usage_filt = usage[
            (date_ranges[i] <= usage.date)
            & (usage.date <= date_ranges[i + 1])
        ].groupby('node_id')
        cores_used.append(usage_filt['cores'].max().sum())
        cores_avail.append(usage_filt['n_cores'].max().sum())
        ram_used.append(usage_filt['ram'].max().sum())
        ram_avail.append(usage_filt['tot_mem'].max().sum())

    hours = (date_ranges - date_ranges[0]).total_seconds()/60/60

    time_df = pd.DataFrame({
        'time': date_ranges,
        'hour': hours,
        'cores_used': cores_used,
        'cores_avail': cores_avail,
        'ram_used': ram_used,
        'ram_avail': ram_avail,
    })
    return time_df


class TextPlotter:
    """
    ASCII plotter like matplotlib but in command line.
    """

    canvas_width: int
    canvas_height: int

    def __init__(
        self, width: Optional[int] = None, height: Optional[int] = None
    ):
        term = shutil.get_terminal_size()
        if height is None:
            self.canvas_height = term.lines - 1
        else:
            self.canvas_height = height
        if width is None:
            self.canvas_width = term.columns - 1
        else:
            self.canvas_width = width
        self._xlim: List[Optional[float]] = [None, None]
        self._ylim: List[Optional[float]] = [None, None]
        self._xlabel: Optional[str] = None
        self._ylabel: Optional[str] = None
        self._data: List[Dict[str, Any]] = []
        self._title: Optional[str] = None
        self._legend: bool = False

    def title(self, label) -> None:
        self._title = label

    def _blank_canvas(self) -> None:
        self.lines = [
            [' ' for j in range(self.canvas_width)]
            for i in range(self.canvas_height)
        ]

    def _draw_axes(self) -> None:
        for j in range(self.plot_height):
            self.lines[self.top_margin + j][self.left_margin] = '|'
            self.lines[self.top_margin + j][
                self.canvas_width - self.right_margin - 1
            ] = '|'
        for j in range(self.plot_width):
            self.lines[
                self.canvas_height - self.bottom_margin
            ][j + self.left_margin] = '-'
            self.lines[
                self.top_margin
            ][j + self.left_margin] = '-'
        self.lines[
            self.canvas_height - self.bottom_margin
        ][self.left_margin] = '+'
        self.lines[
            self.canvas_height - self.bottom_margin
        ][self.canvas_width - self.right_margin - 1] = '+'
        self.lines[self.top_margin][self.left_margin] = '+'
        self.lines[self.top_margin][
            self.canvas_width - self.right_margin - 1
        ] = '+'

    def _init_sizes(self) -> None:
        self.left_margin = 5
        if self._ylabel is not None:
            self.left_margin = min(len(self._ylabel) + 2, 20)

        self.right_margin = 2
        if self._legend:
            legend_rows = self._split_text(self._get_legend_str(), wrap=20)
            legend_width = max(len(row) for row in legend_rows)
            self.right_margin = max(5, legend_width + 2)

        self.bottom_margin = 2
        if self._xlabel:
            self.bottom_margin = 4

        self.plot_width = (
            self.canvas_width - self.left_margin - self.right_margin
        )

        self.top_margin = 1
        if self._title:
            n_title_rows = len(
                self._split_text(self._title, wrap=self.plot_width)
            )
            self.top_margin = max(1, n_title_rows + 1)

        self.plot_height = (
            self.canvas_height - self.bottom_margin - self.top_margin
        )

    def _split_text(self, text: str, wrap: int) -> List[str]:
        split_text = []
        for line in text.split('\n'):
            if len(line) <= wrap:
                split_text.append(line)
            else:
                words = line.split(' ')
                subline = ''
                while len(words) > 0:
                    next_word = words.pop(0)
                    if len(next_word) >= wrap:
                        split_text.append(subline)
                        split_text.append(next_word)
                        subline = ''
                    elif len(subline) + len(next_word) >= wrap:
                        split_text.append(subline)
                        subline = next_word
                    else:
                        if subline:
                            subline += ' ' + next_word
                        else:
                            subline = next_word
                if subline:
                    split_text.append(subline)
        return split_text

    def _label_text(
        self, label: str, x: int, y: int, vertical: bool = False,
        halign: str = 'left', valign: str = 'top',
        wrap: Optional[int] = None
    ) -> None:

        if wrap is None:
            if vertical:
                wrap = self.canvas_height
            else:
                wrap = self.canvas_width

        # Split into rows.
        label_rows = self._split_text(label, wrap=wrap)
        n_rows = len(label_rows)

        voffset = 0
        if n_rows > 1:
            if valign == 'center':
                voffset = int(-len(label_rows)/2)
            elif valign == 'bottom':
                voffset = int(-len(label_rows))

        if vertical:
            if x + voffset <= 0:
                voffset = -x
            elif x + n_rows + voffset >= self.canvas_width:
                voffset = self.canvas_width - x - n_rows
        else:
            if y + voffset <= 0:
                voffset = -y
            elif y + n_rows + voffset >= self.canvas_height:
                voffset = self.canvas_height - y - n_rows

        for i, row in enumerate(label_rows):
            row_width = len(row)

            # Offset depending on alignment.
            hoffset = 0
            if halign == 'right':
                hoffset = -row_width
            elif halign == 'center':
                hoffset = int(-row_width/2)

            if vertical:

                # Ensure text within boundaries.
                if y + hoffset <= 0:
                    hoffset = -y
                elif y + hoffset + row_width >= self.canvas_height:
                    hoffset = self.canvas_height - y - hoffset - row_width

                for j, char in enumerate(row):
                    self.lines[y + j + hoffset][x + i + voffset] = char
            else:

                # Ensure text within boundaries.
                if x + hoffset <= 0:
                    hoffset = -x
                elif x + hoffset + row_width > self.canvas_width:
                    hoffset = self.canvas_width - x - hoffset - row_width - 2

                for j, char in enumerate(row):
                    self.lines[y + i + voffset][j + x + hoffset] = char

    def xlabel(self, label: str) -> None:
        self._xlabel = label

    def ylabel(self, label: str) -> None:
        self._ylabel = label

    def _draw_xlabel(self) -> None:
        if self._xlabel is not None:
            x = int(self.left_margin + self.plot_width/2)
            y = int(self.canvas_height - self.bottom_margin/2)
            self._label_text(self._xlabel, x, y, halign='center')

    def _draw_ylabel(self) -> None:
        if self._ylabel is not None:
            max_width = max(
                len(line)
                for line in self._split_text(
                    self._ylabel, wrap=self.left_margin - 1
                )
            )
            if max_width < self.left_margin:
                x = int(self.left_margin/2)
                y = int(self.top_margin + self.plot_height/2)
                self._label_text(
                    self._ylabel, x, y, vertical=False,
                    halign='center', valign='center',
                    wrap=self.left_margin - 1
                )
            else:
                x = int(self.left_margin/2)
                y = int(self.top_margin + self.plot_height/2)
                self._label_text(
                    self._ylabel, x, y, vertical=True,
                    halign='center', valign='center'
                )

    def fill(
        self, x: np.ndarray, y: np.ndarray, marker: str = '.',
        label: Optional[str] = None
    ) -> None:
        self._data.append({
            'linetype': 'fill',
            'x': x,
            'y': y,
            'marker': marker,
            'label': label,
        })

    def plot(
        self, x: np.ndarray, y: np.ndarray, marker: str = '.',
        label: Optional[str] = None
    ) -> None:
        self._data.append({
            'linetype': 'plot',
            'x': x,
            'y': y,
            'marker': marker,
            'label': label,
        })

    def _get_limits(self) -> Tuple[float, float, float, float]:
        x_lim_min: float = self._xlim[0] if self._xlim[0] is not None else 0
        x_lim_max: float = self._xlim[1] if self._xlim[1] is not None else 1
        y_lim_min: float = self._ylim[0] if self._ylim[0] is not None else 0
        y_lim_max: float = self._ylim[1] if self._ylim[1] is not None else 1
        return x_lim_min, x_lim_max, y_lim_min, y_lim_max

    def _draw_fill(self) -> None:
        x_lim_min, x_lim_max, y_lim_min, y_lim_max = self._get_limits()
        for series in self._data:
            if series['linetype'] == 'fill':
                x = np.array(series['x'])
                y = np.array(series['y'])
                marker = series['marker']
                x_spacing = (x_lim_max - x_lim_min)/self.plot_width
                y_spacing = (y_lim_max - y_lim_min)/self.plot_height
                for i in range(self.plot_width):
                    x_min = x_lim_min + i*x_spacing
                    x_max = x_lim_min + (i + 1)*x_spacing
                    index = np.where((x_min <= x) & (x <= x_max))[0]
                    if len(index) > 0:
                        for j in range(self.plot_height):
                            y_row = y_lim_min + j*y_spacing
                            y_max = np.max(y[index])
                            if y_row <= y_max:
                                self.lines[
                                    self.canvas_height - self.bottom_margin - j
                                ][i + self.left_margin] = marker

    def _draw_plot(self) -> None:
        x_lim_min, x_lim_max, y_lim_min, y_lim_max = self._get_limits()
        for series in self._data:
            if series['linetype'] == 'plot':
                x = np.array(series['x'])
                y = np.array(series['y'])
                marker = series['marker']
                x_spacing = (x_lim_max - x_lim_min)/self.plot_width
                y_spacing = (y_lim_max - y_lim_min)/self.plot_height
                for i in range(self.plot_width):
                    for j in range(self.plot_height):
                        x_min = x_lim_min + i*x_spacing
                        x_max = x_lim_min + (i + 1)*x_spacing
                        y_min = y_lim_min + j*y_spacing
                        y_max = y_lim_min + (j + 1)*y_spacing
                        if np.any(
                            (x_min <= x) & (x <= x_max)
                            & (y_min <= y) & (y <= y_max)
                        ):
                            self.lines[
                                self.canvas_height - self.bottom_margin - j
                            ][i + self.left_margin] = marker

    def legend(self) -> None:
        self._legend = True

    def _format_number(self, value: Optional[Union[int, float]]) -> str:
        if value is None:
            return ''
        elif value % 1 == 0:
            return str(int(value))
        else:
            return '{:.4g}'.format(value)

    def _label_limits(self) -> None:
        self._label_text(
            self._format_number(self._xlim[0]),
            self.left_margin, self.top_margin + self.plot_height + 1,
            halign='left', valign='top'
        )
        self._label_text(
            self._format_number(self._xlim[1]),
            self.canvas_width - self.right_margin,
            self.top_margin + self.plot_height + 1,
            halign='center', valign='top'
        )
        self._label_text(
            self._format_number(self._ylim[0]),
            self.left_margin - 1, self.top_margin + self.plot_height,
            halign='right', valign='bottom'
        )
        self._label_text(
            self._format_number(self._ylim[1]),
            self.left_margin - 1, self.top_margin,
            halign='right', valign='top'
        )

    def _draw_title(self) -> None:
        if self._title is not None:
            self._label_text(
                self._title,
                int(self.left_margin + self.plot_width/2),
                int(self.top_margin/2),
                halign='center', valign='center'
            )

    def _set_limits(self) -> None:
        if self._xlim[0] is None:
            self._xlim[0] = min([np.min(series['x']) for series in self._data])
        if self._xlim[1] is None:
            self._xlim[1] = max([np.max(series['x']) for series in self._data])
        if self._ylim[0] is None:
            self._ylim[0] = min([np.min(series['y']) for series in self._data])
        if self._ylim[1] is None:
            self._ylim[1] = max([np.max(series['y']) for series in self._data])

    def ylim(self, y_min: float, y_max: float) -> None:
        self._ylim = [y_min, y_max]

    def xlim(self, x_min: float, x_max: float) -> None:
        self._xlim = [x_min, x_max]

    def _get_legend_str(self) -> str:
        legend_strings = []
        for series in self._data:
            if 'label' in series and series['label'] is not None:
                label = series['label']
                marker = series['marker']
                legend_strings.append(marker*3 + ' ' + label)
        return '\n'.join(legend_strings)

    def _draw_legend(self) -> None:
        if self._legend:
            self._label_text(
                self._get_legend_str(),
                self.canvas_width - self.right_margin + 2,
                int(self.canvas_height/2),
                halign='left', valign='center',
                wrap=self.right_margin
            )

    def render(self) -> str:
        self._init_sizes()
        self._blank_canvas()
        self._set_limits()
        self._draw_fill()
        self._draw_plot()
        self._draw_axes()
        self._draw_xlabel()
        self._draw_ylabel()
        self._draw_title()
        self._draw_legend()
        self._label_limits()
        return '\n'.join([''.join(line) for line in self.lines])

    def show(self) -> None:
        render = self.render()
        print(render)



import os
import json
import pytest
import gzip
import numpy as np
from panqec.error_models import PauliErrorModel
from panqec.codes import Toric2DCode
from panqec.decoders import BeliefPropagationOSDDecoder
from panqec.simulation import (
    read_input_json, run_once, DirectSimulation, expand_input_ranges, run_file,
    BatchSimulation
)
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(BASE_DIR, 'data')


@pytest.fixture
def required_fields():
    return [
        'effective_error', 'success'
    ]


@pytest.mark.parametrize(
    'file_name, expected_runs',
    [
        ('single_input.json', 1),
        ('range_input.json', 27),
        ('single_input.json.gz', 1),
        ('range_input.json.gz', 27),
    ]
)
def test_read_json_input(file_name, expected_runs):
    input_json = os.path.join(DATA_DIR, file_name)
    output_json = os.path.join(DATA_DIR, 'results.json')
    batch_simulation = read_input_json(input_json, output_json)
    assert batch_simulation is not None
    assert len(batch_simulation._simulations) == expected_runs
    parameters = [
        {
            'code': list(s.code.size),
            'error_model': s.error_model.direction,
            'error_rate': s.error_rate
        }
        for s in batch_simulation._simulations
    ]
    for i in range(len(parameters)):
        for j in range(len(parameters)):
            if i != j:
                assert parameters[i] != parameters[j]


class TestRunOnce:
    @pytest.fixture
    def code(self):
        return Toric2DCode(3, 3)

    @pytest.fixture
    def error_model(self):
        return PauliErrorModel(1/3, 1/3, 1/3)

    def test_run_once(self, required_fields, code, error_model):
        error_rate = 0.5
        decoder = BeliefPropagationOSDDecoder(code, error_model, error_rate)
        results = run_once(
            code, error_model, decoder, error_rate=error_rate
        )
        assert set(required_fields).issubset(results.keys())
        assert results['error'].shape == (2*code.n,)
        assert results['syndrome'].shape == (code.stabilizer_matrix.shape[0],)
        assert isinstance(results['success'], bool)
        assert results['effective_error'].shape == (
            2*code.logicals_x.shape[0],
        )
        assert isinstance(results['codespace'], bool)

    def test_run_once_invalid_probability(self, code, error_model):
        error_rate = -1
        decoder = BeliefPropagationOSDDecoder(
            code, error_model, error_rate=error_rate
        )
        with pytest.raises(ValueError):
            run_once(code, error_model, decoder, error_rate=error_rate)


class TestSimulationToric2DCode():

    error_rate = 0.5

    @pytest.fixture
    def code(self):
        return Toric2DCode(3, 3)

    @pytest.fixture
    def decoder(self, code, error_model):
        return BeliefPropagationOSDDecoder(code, error_model, self.error_rate)

    @pytest.fixture
    def error_model(self):
        direction = np.ones(3)/3
        return PauliErrorModel(*direction)

    def test_run(self, code, error_model, decoder, required_fields):
        error_rate = 0.5
        simulation = DirectSimulation(code, error_model, decoder, error_rate)
        simulation.run(10)
        assert len(simulation._results['success']) == 10
        assert set(required_fields).issubset(simulation._results.keys())


class TestBatchSimulationOneFile():

    n_trials: int = 5

    @pytest.fixture
    def output_file(self, tmpdir):
        out_file = os.path.join(tmpdir, 'results.json.gz')

        batch_sim = BatchSimulation(
            label='mylabel',
            save_frequency=1,
            output_file=out_file,
        )

        for size, error_rate in [(3, 0.1), (4, 0.5)]:
            code = Toric2DCode(size, size)
            error_model = PauliErrorModel(1/3, 1/3, 1/3)
            decoder = BeliefPropagationOSDDecoder(
                code, error_model, error_rate
            )
            simulation = DirectSimulation(
                code, error_model, decoder, error_rate
            )
            batch_sim.append(simulation)

        assert len(batch_sim) == 2

        batch_sim.run(self.n_trials)
        batch_sim.save_results()

        return out_file

    def test_output_to_one_file(self, output_file):
        assert os.path.isfile(output_file)

        with gzip.open(output_file, 'rb') as gz:
            results = json.loads(gz.read().decode('utf-8'))

        # Check integrity of results.
        assert len(results) == 2
        expected_input_keys = [
            'code', 'error_model', 'decoder',
            'error_rate', 'method'
        ]
        expected_results_keys = [
            'effective_error', 'success', 'codespace', 'wall_time',
        ]
        for key in expected_input_keys:
            assert key in results[0]['inputs']
            assert key in results[1]['inputs']

        for key in expected_results_keys:
            assert key in results[0]['results']
            assert key in results[1]['results']

        # Numerical tests on the results.
        assert len(results[0]['results']['effective_error']) == self.n_trials
        assert len(results[0]['results']['success']) == self.n_trials
        assert len(results[0]['results']['codespace']) == self.n_trials


@pytest.fixture
def example_ranges():
    input_json = os.path.join(DATA_DIR, 'range_input.json')
    with open(input_json) as f:
        data = json.load(f)
    return data['ranges']


def test_expand_inputs_ranges(example_ranges):
    ranges = example_ranges
    expanded_inputs = expand_input_ranges(ranges)
    assert len(expanded_inputs) == 27


def test_run_file_range_input(tmpdir):
    input_json = os.path.join(DATA_DIR, 'range_input.json')
    output_json = os.path.join(tmpdir, 'results', 'results.json')
    n_trials = 2
    run_file(input_json, output_json, n_trials)

    assert os.path.isfile(output_json)


class TestReadInputJson:

    def test_multiple_ranges(self):
        input_json = os.path.join(DATA_DIR, 'toric_input.json')
        output_json = os.path.join(DATA_DIR, 'toric_output.json')

        batch_sim = read_input_json(input_json, output_json)
        assert len(batch_sim) == 126



from typing import Tuple, Dict, List
import numpy as np
from panqec.codes import StabilizerCode

Operator = Dict[Tuple, str]  # Location to pauli ('X','Y','Z')
Coordinates = List[Tuple]  # List of locations


class Planar3DCode(StabilizerCode):
    """3D surface code on cubic lattice with open boundary conditions.

    Parameters
    ----------
    L_x : int
        Number of edges along the x direction (on edges parallel to x).
    L_y : Optional[int]
        Number of qubits in the y direction (on edges parallel to x).
    L_z : Optional[int]
        Number of qubits in the z direction (on edges parallel to x).

    Notes
    -----
    Qubits live on edges.
    Boundaries parallel to the xy plane (i.e. orthogonal to z) are smooth.
    Boundaries parallel to the zx plane (i.e. orthogonal to y) are rough.
    Boundaries parallel to the yz plane (i.e. orthogonal to x) are smooth.
    Similar to :class:`panqec.codes.surface_2d.Planar2DCode`
    but extruded in z direction with smooth open boundaries and vertical z
    edges (with qubits) in between each layers' vertices.
    Stabilizers are otherwise the same as that of
    :class:`panqec.codes.surface_3d.Toric3DCode`
    but truncated at the boundaries.
    """
    dimension = 3
    deformation_names = ['XZZX']

    @property
    def label(self) -> str:
        return 'Planar {}x{}x{}'.format(*self.size)

    def get_qubit_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        # Qubits along e_x
        for x in range(1, 2*Lx+1, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(0, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Qubits along e_y
        for x in range(2, 2*Lx, 2):
            for y in range(1, 2*Ly-1, 2):
                for z in range(0, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Qubits along e_z
        for x in range(2, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(1, 2*Lz-1, 2):
                    coordinates.append((x, y, z))

        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        # Vertices
        for x in range(2, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(0, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Faces in xy plane
        for x in range(1, 2*Lx+1, 2):
            for y in range(1, 2*Ly-1, 2):
                for z in range(0, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Faces in yz plane
        for x in range(2, 2*Lx, 2):
            for y in range(1, 2*Ly-1, 2):
                for z in range(1, 2*Lz-1, 2):
                    coordinates.append((x, y, z))

        # Faces in xz plane
        for x in range(1, 2*Lx+1, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(1, 2*Lz-1, 2):
                    coordinates.append((x, y, z))

        return coordinates

    def stabilizer_type(self, location: Tuple) -> str:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location} for a stabilizer")

        x, y, z = location
        if x % 2 == 0 and y % 2 == 0:
            return 'vertex'
        else:
            return 'face'

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location} for a stabilizer")

        if self.stabilizer_type(location) == 'vertex':
            pauli = 'Z'
        else:
            pauli = 'X'

        x, y, z = location

        if self.stabilizer_type(location) == 'vertex':
            delta = [(1, 0, 0), (-1, 0, 0), (0, 1, 0), (0, -1, 0), (0, 0, 1),
                     (0, 0, -1)]
        else:
            # Face in xy-plane.
            if z % 2 == 0:
                delta = [(-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0)]
            # Face in yz-plane.
            elif (x % 2 == 0):
                delta = [(0, -1, 0), (0, 1, 0), (0, 0, -1), (0, 0, 1)]
            # Face in zx-plane.
            elif (y % 2 == 0):
                delta = [(-1, 0, 0), (1, 0, 0), (0, 0, -1), (0, 0, 1)]

        operator = dict()
        for d in delta:
            qubit_location = tuple(np.add(location, d))

            if self.is_qubit(qubit_location):
                operator[qubit_location] = pauli

        return operator

    def qubit_axis(self, location) -> str:
        x, y, z = location

        if (z % 2 == 0) and (x % 2 == 1) and (y % 2 == 0):
            axis = 'x'
        elif (z % 2 == 0) and (x % 2 == 0) and (y % 2 == 1):
            axis = 'y'
        elif (z % 2 == 1) and (x % 2 == 0) and (y % 2 == 0):
            axis = 'z'
        else:
            raise ValueError(f'Location {location} does not correspond'
                             'to a qubit')

        return axis

    def get_logicals_x(self) -> List[Operator]:
        """The unique logical X operator."""

        Lx, Ly, Lz = self.size
        logicals = []

        # X operators along x edges in x direction.
        operator: Operator = dict()
        for x in range(1, 2*Lx+1, 2):
            operator[(x, 0, 0)] = 'X'
        logicals.append(operator)

        return logicals

    def get_logicals_z(self) -> List[Operator]:
        """The unique logical Z operator."""

        Lx, Ly, Lz = self.size
        logicals = []

        # X operators along x edges in x direction.
        operator: Operator = dict()
        for y in range(0, 2*Ly, 2):
            for z in range(0, 2*Lz, 2):
                operator[(1, y, z)] = 'Z'
        logicals.append(operator)

        return logicals

    def stabilizer_representation(
        self, location, rotated_picture=False, json_file=None
    ) -> Dict:
        representation = super().stabilizer_representation(
            location, rotated_picture, json_file=json_file
        )

        x, y, z = location
        if not rotated_picture and self.stabilizer_type(location) == 'face':
            if z % 2 == 0:  # xy plane
                representation['params']['normal'] = [0, 0, 1]
            elif x % 2 == 0:  # yz plane
                representation['params']['normal'] = [1, 0, 0]
            else:  # xz plane
                representation['params']['normal'] = [0, 1, 0]

        if rotated_picture and self.stabilizer_type(location) == 'face':
            if z % 2 == 0:
                representation['params']['normal'] = [0, 0, 1]
            elif x % 2 == 0:
                representation['params']['normal'] = [1, 0, 0]
            else:
                representation['params']['normal'] = [0, 1, 0]

        return representation

    def get_deformation(
        self, location: Tuple,
        deformation_name: str,
        deformation_axis: str = 'z',
        **kwargs
    ) -> Dict:

        if deformation_axis not in ['x', 'y', 'z']:
            raise ValueError(f"{deformation_axis} is not a valid "
                             "deformation axis")

        if deformation_name == 'XZZX':
            undeformed_dict = {'X': 'X', 'Y': 'Y', 'Z': 'Z'}
            deformed_dict = {'X': 'Z', 'Y': 'Y', 'Z': 'X'}

            if self.qubit_axis(location) == deformation_axis:
                deformation = deformed_dict
            else:
                deformation = undeformed_dict

        else:
            raise ValueError(f"The deformation {deformation_name}"
                             "does not exist")

        return deformation



import itertools
import pytest
import numpy as np
from panqec.bpauli import bsf_wt
from panqec.codes import Toric3DCode
from panqec.decoders import MatchingDecoder
from panqec.error_models import PauliErrorModel
from tests.decoders.decoder_test import DecoderTest


class TestToric3DMatchingDecoder(DecoderTest):

    @pytest.fixture
    def code(self):
        return Toric3DCode(3, 4, 5)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.1
        return MatchingDecoder(code, error_model, error_rate, 'X')

    @pytest.fixture
    def allowed_paulis(self):
        return ['X']

    def test_decode_X_error(self, decoder, code):
        error = code.to_bsf({
            (2, 1, 2): 'X',
        })
        assert bsf_wt(error) == 1

        # Measure the syndrome and ensure non-triviality.
        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2

        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_decode_many_X_errors(self, decoder, code):
        error = code.to_bsf({
            (1, 0, 0): 'X',
            (0, 1, 0): 'X',
            (0, 0, 3): 'X',
        })
        assert bsf_wt(error) == 3

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_unable_to_decode_Z_error(self, decoder, code):
        error = code.to_bsf({
            (1, 0, 2): 'Z'
        })

        assert bsf_wt(error) == 1

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        assert np.all(correction == 0)

        total_error = (error + correction) % 2
        assert np.all(error == total_error)

        assert np.any(code.measure_syndrome(total_error) != 0)

    def test_decode_many_codes_and_errors_with_same_decoder(self):

        codes = [
            Toric3DCode(3, 4, 5),
            Toric3DCode(3, 3, 3),
            Toric3DCode(5, 4, 3),
        ]

        sites = [
            (0, 0, 1),
            (1, 0, 0),
            (0, 1, 0)
        ]

        error_model = PauliErrorModel(1/3, 1/3, 1/3)
        error_rate = 0.1

        for code, site in itertools.product(codes, sites):
            decoder = MatchingDecoder(code, error_model, error_rate, 'X')
            error = code.to_bsf({
                site: 'X',
            })
            syndrome = code.measure_syndrome(error)
            correction = decoder.decode(syndrome)
            total_error = (error + correction) % 2
            assert np.all(code.measure_syndrome(total_error) == 0)
